# -*- coding: utf8 -*-
"""Sphinx extension for autodocumenting fortran codes.


"""
# Copyright or © or Copr. Actimar (contributor(s) : Stephane Raynaud) (2010)
# 
# raynaud@actimar.fr
# 
# 
# This software is a computer program whose purpose is to provide
# utilities for handling oceanographic and atmospheric data,
# with the ultimate goal of validating the MARS model from IFREMER.
# 
# This software is governed by the CeCILL license under French law and
# abiding by the rules of distribution of free software.  You can  use, 
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info". 
# 
# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability. 
# 
# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or 
# data to be ensured and,  more generally, to use and operate it in the 
# same conditions as regards security. 
# 
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.
# 
from sphinx.directives import Directive
from docutils.parsers.rst.directives import unchanged
from docutils import nodes
from docutils.statemachine import string2lines
from glob import glob
from numpy.f2py.crackfortran import crackfortran, fortrantypes
import re, os, sys
from fortran_domain import FortranDomain


# Fortran parser and formatter
# ----------------------------

class F90toRstException(Exception):
    pass

class F90toRst(object):
    '''Fortran 90 parser and restructeredtext formatter
    
    :Parameters:
    
         - **ffiles**: Fortran files (glob expression allowed) or dir (or list of)
         
    :Options:
    
        - **ic**: Indentation char.
        - **ulc**: Underline char for titles.
        - **sst**: Subsection type.
        - **vl**: Verbose level (0=quiet).
    '''
    _re_unended_match = re.compile(r'(.*)&\s*',re.I).match
    _re_unstarted_match = re.compile(r'\s*&(.*)',re.I).match
    _re_comment_match = re.compile(r'\s*!(.*)',re.I).match
    _re_space_prefix_match = re.compile(r'^(\s*).*$',re.I).match
#    _re_sub_numvardesc_match = re.compile(r'\$(\d+)',re.I).match
#    _re_vardesc_match = re.compile(r'\s*(?P<varname>\w+)[\s,&]*!\s*(?P<vardesc>.+)',re.I).match
#    _re_type_var_findall = re.compile(r'(\w+)',re.I).match
    _fmt_vardesc = ':%(role)s %(vtype)s %(vname)s%(vdim)s%(vattr)s: %(vdesc)s'
    #_fmt_vardesc = '**%(vname)s** %(vdim)s :: %(vtype)s %(vattr)s: %(vdesc)s'
    _fmt_vattr = " [%(vattr)s]"
    _fmt_fvardesc = '%(vtype)s%(vdim)s %(vattr)s%(vdesc)s'
    #_fmt_fvardim = ', **dim%(vdim)s**'
#    _re_call_subr_search = re.compile(r'\bcall\s+(\w+)\(',re.I).search
#    _re_call_func_findall = re.compile(r'\b(\w+)\(',re.I).search


    def __init__(self, ffiles, ic='\t', ulc='-', vl=0, encoding='utf8', sst='rubric'):
        # Be sure to have a list
        if not isinstance(ffiles, list):
            ffiles = [ffiles]
        
        # Read and store them
        self.src = {}
        self.ffiles = ffiles
        for ff in ffiles:
            f = open(ff)
            self.src[ff] = []
            for l in f.readlines():
                try:
                    self.src[ff].append(l[:-1].decode(encoding) )
                except:
                    raise F90toRstException('Encoding error\n  file = %s\n  line = %s'%(ff,l))
            #self.src[ff] = [l.decode(encoding) for l in f.readlines()]
            #self.src[ff] = [l.decode(encoding) for l in f.read().split('\n')]
           
            f.close()

        # Crack files
        global verbose, quiet
        import numpy.f2py.crackfortran
        self._verbose = numpy.f2py.crackfortran.verbose = verbose = vl
        numpy.f2py.crackfortran.quiet = quiet = 1-verbose
        self.crack = crackfortran(ffiles)
        
        # Build index
        self.build_index()
        
        # Scan all files to get description, etc
        self.scan()
        
        # Add function 'call from' to index
        self.build_callfrom_index()
        
        # Other inits
        self.rst = {}
        self._ic = ic
        self._ulc = ulc
        self._sst = sst
        
    # Indexing ---
        
    def build_index(self):
        """Register modules, functions, types and module variables for quick access
        
        Index constituents are:
        
            .. attribute:: modules
            
                Dictionary where each key is a module name, and each value is the cracked block.
                
            .. attribute:: routines
                   
                Module specific functions and subroutines
                
            .. attribute:: types
            
                Module specific types
            
            .. attribute:: variables
        
                Module specific variables
        """
        # Containers
        self.modules = {}
        self.types = {}
        self.routines = self.functions = self.subroutines = {}
        self.variables = {}
        self.programs = {}
        
        # Loop on all blocks and subblocks
        for block in self.crack:
            
            # Modules
            if block['block']=='module':
                           
                # Index module                
                module = block['name']
                self.modules[module] = block
                
                # Loop inside module
                for subblock in block['body']:

                    # Types and routines as subblock
                    if subblock['block'] in ['function', 'type', 'subroutine']:
                        
                        # Index
                        container = getattr(self, subblock['block']+'s')
                        container[subblock['name']] = subblock
                        subblock['module'] = module
                      
                        # Variables
                        for varname, bvar in subblock['vars'].items():
                            bvar['name'] = varname
                            
                # Function aliases from "use only" (rescan)
                for bfunc in self.routines.values(): bfunc['aliases'] = []
                for subblock in block['body']:
                    if not subblock['block'] == 'use': continue
                    for monly in block['use'].values():
                        if not monly: continue
                        for fname, falias in monly['map'].items():
                            self.routines[falias] = self.routines[fname]
                            if falias not in self.routines[fname]['aliases']:
                                self.routines[fname]['aliases'].append(falias)
                       
                # Module variables
                for varname, bvar in block['vars'].items():
                    self.variables[varname] = bvar
                    bvar['name'] = varname
                    bvar['module'] = module
                    
                
            # Local functions, subroutines and programs
            elif block['block'] in ['function', 'subroutine', 'program']:
                
                # Index
                container = getattr(self, block['block']+'s')
                container[block['name']] = block
              
                # Variables
                for varname, bvar in block['vars'].items():
                    bvar['name'] = varname
        
                    
        # Regular expression for fast search
        # - function calls
        subs = [block['name'].lower() for block in self.routines.values() if block['block']=='subroutine']
        self._re_callsub_findall = subs and \
            re.compile(r'call\s+(%s)\b'%('|'.join(subs)), re.I).findall or (lambda line: [])
        funcs = [block['name'].lower() for block in self.routines.values() if block['block']=='function']
        self._re_callfunc_findall = funcs and \
            re.compile(r'\b(%s)\s*\('%('|'.join(funcs)), re.I).findall or (lambda line: [])
        # - function variables
        for block in self.routines.values():
            vars = (r'|\b').join(block['sortvars']) + r'|\$(?P<varnum>\d+)'
            sreg = r'[\s\*\-:]*(?:@param\s*)?(?P<varname>\b%s\b)\W+(?P<vardesc>.*)'%vars
            block['vardescmatch'] = re.compile(sreg).match
        # - variables with description
#        for block in self.types.values()+self.modules.values():
        for block in self.types.values()+self.modules.values()+self.routines.values():
            sreg = r'\b(?P<varname>%s)\b[\W\d]*!\s*(?P<vardesc>.*)'%'|'.join(block['sortvars'])
            if block['sortvars']:
                block['vardescsearch'] = re.compile(sreg).search
            else:
                block['vardescsearch'] = lambda x: None

        

    def build_callfrom_index(self):
        """For each function, index which function call it"""
        for bfunc in self.routines.values():
            bfunc['callfrom'] = []
            for bfuncall in self.routines.values()+self.programs.values():
                if bfunc['name'] in bfuncall['callto']:
                    bfunc['callfrom'].append(bfuncall['name'])
                    
    def filter_by_srcfile(self, sfile, mode='basename', objtype=None):
        """Search for subblocks according to origin file
        
        :Params:
            - **sfile**: Source file name.
            - **mode**, optional: Mode for searching for sfile.
              If ``"strict"``, exact match is needed, else only basename.
            - **objtype**, optional: Restrict search to one or a list of object types
              (i.e. ``"function"``, ``"program"``, etc).
        """
        if blocktype and not isinstance(blocktype, list):
            blocktype = [blocktype]
        bb = []
        if mode!='strict': 
            sfile = os.path.basename(sfile)
        for b in self.crack:
            if objtype and objtype != 'all' and block['block'] not in objtype: continue
            bfile = b['from'].split(':')[0] # remove module name
            if mode=='strict':
                bfile = os.path.basename(bfile)
            if sfile==bfile: bb.append(b)
        return bb
         
    def scan(self):
        """Scan """
        # Loop on all blocks
        for block in self.crack:
            
            # Modules
            if block['block']=='module':

                # Get source lines
                modsrc = self.get_blocksrc(block)

                # Get module description
                block['desc'] = self.get_comment(modsrc, aslist=True)
                
                # Scan types and routines
                for subblock in block['body']:

                    self.scan_container(subblock, insrc=modsrc)
                    
                # Scan module variables
                self.strip_blocksrc(block, ['type', 'function', 'subroutine'], src=modsrc)
                for line in modsrc:
                    if line.strip().startswith('!'): continue
                    m = block['vardescsearch'](line)
                    if m:
                        block['vars'][m.group('varname')]['desc'] = m.group('vardesc')
                for bvar in block['vars'].values():
                    bvar.setdefault('desc', '')
            
            # Routines
            elif block['block'] in ['function', 'subroutine', 'program']:
                
                self.scan_container(block)


    def scan_container(self, block, insrc=None):
        """Scan a block of program, routines or type"""
        
        # Check block type
        if block['block'] not in ['type', 'function', 'subroutine', 'program']: return
                
        # Source lines
        subsrc = self.get_blocksrc(block, insrc)
        
        # Comment
        block['desc'] = self.get_comment(subsrc, aslist=True)
        
        # Callable objects
        if block['block'] in ['function', 'subroutine', 'program']: 
            
            # With signature : description of variables in comment
            if block['block'] in ['function', 'subroutine']:
                varname = None
                for iline, line in enumerate(block['desc']):
                    m = block['vardescmatch'](line)
                    if m: # There is a variable and its description
                    
                        # Name of variable
                        varname = m.group('varname')
                        
                        # Numeric name
                        if m.group('varnum'): # $[1-9]+
                            ivar = int(m.group('varnum'))
                            ivar = ivar-1
                            if ivar<0 or ivar>=len(block['args']): continue
                            block['desc'][iline] = line.replace(varname, block['args'][ivar])
                            varname = block['args'][ivar]
                            
                        # Store description
                        ifirst = len(line)-len(line.strip())
                        block['vars'][varname]['desc'] = m.group('vardesc')
                        
                    elif line.strip() and varname is not None and \
                        (len(line)-len(line.strip()))>ifirst: # Description continuation?
                        
                       block['vars'][varname]['desc'].append(' '+line.strip())
                       
                    else: varname = None
                                        
            
            # Index calls
            block['callto'] = []
            if subsrc is not None:
                self.join_src(subsrc)
                for line in subsrc[1:-1]:
                    if line.strip().startswith('!'): continue
                    line = line.lower()
                    for fn in self._re_callsub_findall(line)+self._re_callfunc_findall(line):
                        if fn not in block['callto']:
                            block['callto'].append(fn)
                            pass
        
                
            
        # Get description of variables (overwrite for functions and subroutines)
        if block['block'] in ['function', 'subroutine'] and subsrc is not None:
            for line in subsrc:
                if line.strip().startswith('!'): continue
                m = block['vardescsearch'](line)
                if m:
                    block['vars'][m.group('varname')]['desc'] = m.group('vardesc')
                
        # Fill empty descriptions
        for bvar in block['vars'].values():
            bvar.setdefault('desc', '')
                    
        del subsrc
        
                    
    # Getting info ---        

    def get_module(self, block): 
        """Get the name of the current module"""          
        while block['block']!='module':
            if block['parentblock']=='unknown':
                break
            block = block['parentblock']
        return block['name']
           
                
       
    def get_src(self, block):
        """Get the source lines of the file including this block"""
        srcfile = block['from'].split(':')[0]
        return self.src[srcfile]
    
    def join_src(self, src):
        """Join unended lines that does not finish with a comment"""
        for iline, line in enumerate(src):
            m = self._re_unended_match(line)
            if m:
                thisline = m.group(1)
                m = self._re_unstarted_match(src[iline+1])
                nextline = m.group(1) if m else src[iline+1]
                src[iline] = thisline+nextline
                del src[iline+1]
        return src
        
    def get_blocksrc(self, block, src=None, istart=0, getidx=False, stopmatch=None, exclude=None):
        """Extract an identified block of source code
        
        :Parameters:
        
            - *block*: Cracked block
            
        :Options:
        
            - *src*: List of source line including the block
            - *istart*: Start searching from this line number
            
        :Return:
        
            ``None`` or a list of lines            
        """
        # Set up
        if src is None: src = self.get_src(block)
        blocktype = block['block'].lower()
        blockname = block['name'].lower()
        ftypes = '(?:(?:%s).*\s+)?'%fortrantypes if blocktype=='function' else ''
        rstart = re.compile(r"^\s*%s%s\s+%s\b.*$"%(ftypes, blocktype, blockname), re.I).match
        rend   = re.compile(r"^\s*end\s+%s\b.*$"%blocktype, re.I).match
        if isinstance(stopmatch, str):
            stopmatch = re.compile(stopmatch).match
        
        # Beginning
        for ifirst in xrange(istart, len(src)):
            # Simple stop on match
            if stopmatch and stopmatch(src[ifirst]): return
            # Ok, now check
            if rstart(src[ifirst]): break
        else:
            return
        
        # End
        for ilast in xrange(ifirst, len(src)):
            if stopmatch and stopmatch(src[ilast]): break
            if rend(src[ilast].lower()): break
        
        # Extraction
        mysrc = list(src[ifirst:ilast+1])
            
        # Block exclusions
        self.strip_blocksrc(block, exclude, src=mysrc)
            
        if getidx: return mysrc, (ifirst, ilast)
        return mysrc
                      
    def strip_blocksrc(self, block, exc, src=None):
        """Strip blocks from source lines
        
        :Parameters:
        
            - *block*: 
            - *exc* list of block type to remove
        
        :Options:
        
            - *src*: list of source lines        
        
        :Example:
        
        >>> obj.strip_blocksrc(lines, 'type')
        >>> obj.strip_blocksrc(lines, ['function', 'type']
        """
        if src is None: self.get_blocksrc(block)
        if exc is None: return
        if not isinstance(exc, list): exc = [exc]
        for subblock in block['body']:
            if subblock['block'] in exc:
                subsrc = self.get_blocksrc(subblock, src=src, getidx=True)
                if subsrc is None: continue # Already stripped
                del src[subsrc[1][0]:subsrc[1][1]]
                del subsrc 
                
                
    def get_comment(self, src, iline=1, aslist=False, stripped=False, getilast=False, rightafter=True):
        """Search for and return the comment starting after ``iline`` in ``src``
        
        :Params:
        
            - **src**: A list of lines.
            - **iline**, optional: Index of first line to read.
            - **aslist**, optional: Return the comment as a list.
            - **stripped**, optional: Strip each line of comment.
            - **getilast**, optional: Get also index of last line of comment.
            - **rightafter**, optional: Suppose the comment right after 
              the signature line. If True, it prevents from reading a comment
              that is not a description of the routine.
        
        :Return:
       
            - ``scomment``: string or list
            - OR ``scomment,ilast``: if ``getilast is True``
        """
        scomment = []
        in_a_breaked_line = False
        if src is not None:
            for iline in xrange(iline, len(src)):
                line = src[iline].strip()
                
                # Breaked line
                if line.startswith('&'): continue
                
                # Manage no comment line
                m = self._re_comment_match(line)
                if m is None:
                    
                    if not scomment:
                        
                        # Not the end of signature
                        if line.endswith('&'):
                            in_a_breaked_line = True
                            continue
                        if in_a_breaked_line:
                            in_a_breaked_line = False
                            continue
                            
                        # Empty line but we continue searching
                        if not rightafter and not line: 
                            continue
                    
                    # Stop searching
                    break
                
                # Get comment part
                comment = m.group(1)
                
                # Strip?
                if stripped: comment = comment.strip()
                
                # Load and remove space prefix
                if not scomment:
                    prefix = self._re_space_prefix_match(comment).group(1)
                if comment.startswith(prefix): comment = comment[len(prefix):]
                
                # Save comment
                scomment.append(comment)
            
        if not aslist: 
            scomment = self.format_lines(scomment, nlc=' ')
        if getilast: return scomment, iline
        return scomment
        
        
    def get_synopsis(self, block, nmax=2):
        """Get the first ``nmax`` non empty lines of the function, type or module comment as 1 line.
        
        If the header has more than ``nmax`` lines, the first one is taken and appended of '...'.
        If description if empty, it returns an empty string.
        """
        sd = []
        for line in block['desc']:
            line = line.strip()
            if not line:
                if not sd: continue
                break
            sd.append(line)
            if len(sd)>nmax:
                if sd[-1].endswith('.'): sd[-1] += '...'
                break
        if not sd: return ''
        sd = ' '.join(sd)
        return sd
   
    def get_blocklist(self, choice, module):
        """Get the list of types, variables or function of a module"""
        choice = choice.lower()
        if not choice.endswith('s'): choice += 's'
        assert choice in ['types', 'variables', 'functions', 'subroutines'], "Wrong type of declaration"
        module = module.lower()
        assert module in self.modules.keys(), "Wrong module name"
        baselist = getattr(self, choice).values()
        return [v for v in baselist if v.has_key('module') and v['module'] == module.lower()]
        
    # Formating ---
    
    def set_ulc(self, ulc):
        """Set the underline character for title inside module description"""
        self._ulc = ulc
    def get_ulc(self):
        """Get the underline character for title inside module description"""
        return self._ulc
    ulc = property(get_ulc, set_ulc, doc='Underline character for title inside module description')
    
    def set_ic(self, ic):
        """Set the indentation character"""
        self._ic = ic
    def get_ic(self):
        """Get the indentation character"""
        return self._ic
    ic = property(get_ic, set_ic, doc='Indentation character')
   
    def set_sst(self, sst):
        """Set the subsection type"""
        self._sst = sst
    def get_sst(self):
        """Get the subsection type"""
        return self._sst
    sst = property(get_sst, set_sst, doc='Subsection type ("title" or "rubric")')
    
    def indent(self, n):
        """Get a proper indentation"""
        return n*self.ic
    
    def format_lines(self, lines, indent=0, bullet=None, nlc='\n', strip=False):
        """Convert a list of lines to text"""
        if not lines: return ''
        
        # Bullet
        if bullet is True: bullet='-'
        bullet = (str(bullet)+' ') if bullet else ''
        
        # Get current indentation for reduction
        if isinstance(lines, basestring): lines = [lines]

        # Split lines
        tmp = []
        for line in lines:
            if not line:
                tmp.append(line)
            else:
                tmp.extend(line.splitlines())
        lines = tmp ; del tmp
        
        # Expand tabs
        lines = [line.expandtabs(4) for line in lines]

        # Strip
        if strip:
            tmp = []
            for line in lines:
                if not tmp and not line:
                    continue
                tmp.append(line)
            lines = tmp ; del tmp
        if not lines: return ''

        # Column of first non space car
        goodlines = [(len(line)-len(line.lstrip())) for line in lines if line.expandtabs().strip()]
        firstchar = goodlines and min(goodlines) or 0 ; del goodlines
        
        # Redent
        mylines = [self.indent(indent)+bullet+line[firstchar:] for line in lines]
        
        # Create text block
        text = nlc.join(mylines)+nlc
        del mylines
        return text
        

    def format_title(self, text, ulc=None, indent=0):
        """Create a simple rst titlec with indentation
        
        :Parameters:
        
            - *text*: text of the title
        
        :Options:
        
            - *ulc*: underline character (default to attribute :attr:`ucl`)
        
        :Example:
        
            >>> print o.format_title('My title', '-')
            My title
            --------
        """
        if ulc is None: ulc = self.ulc
        return self.format_lines([text, ulc*len(text)], indent=indent)+'\n'
   
    def format_rubric(self, text, indent=0):
        """Create a simple rst rubric with indentation
        
        :Parameters:
        
            - *text*: text of the rubric
        
        :Example:
        
            >>> print o.format_rubric('My title', '-')
            .. rubric:: My rubric
        """
        return self.format_lines('.. rubric:: '+text, indent=indent)+'\n'
        
    def format_subsection(self, text, indent=indent, **kwargs):
        """Format a subsection for describing list of subroutines, types, etc"""
        if self.sst=='title':
            return self.format_title(text, indent=indent, **kwargs)
        return self.format_rubric(text, indent=indent, **kwargs)
    
    def format_declaration(self, dectype, name, description=None, indent=0, bullet=None, options=None):
        """Create an simple rst declaration
        
        :Example:
        
        >>> print format_declaration('var', 'myvar', 'my description', indent=1, bullet='-')
            - .. f:var:: myvar
            
                my description
        """
        declaration = self.format_lines('.. f:%(dectype)s:: %(name)s'%locals(), bullet=bullet, indent=indent)
        if options:
            declaration += self.format_options(options, indent=indent+1)
        declaration += '\n'
        if description:
            declaration += self.format_lines(description, indent=indent+1)
        return declaration+'\n'
        
    def format_options(self, options, indent=0):
        """Format directive options"""
        options = [':%s: %s'%option for option in options.items() if option[1] is not None]
        return self.format_lines(options, indent=indent)

    def format_funcref(self, fname, current_module=None, aliasof=None, module=None):
        """Format the reference link to a module function
        
        Formatting may vary depending on if function is local
        and is an alias.
        
        :Example:
        
        >>> print obj.format_type('myfunc')
        :f:func:`~mymodule.myfunc`
        """            
        # Alias?
        fname = fname.lower()
        if aliasof is not None:
            falias = fname
            fname = aliasof
        if fname in self.routines.keys() and fname in self.routines[fname]['aliases']:
            falias = fname
            fname = self.routines[fname]['name']
        else:
            falias = None
            
        # Local reference ?
        if module is None and self.routines.has_key(fname):
            module = self.routines[fname].get('module')
        if module is None or (current_module is not None and module==current_module):
            if falias:
                return ':f:func:`%(falias)s<%(fname)s>`'%locals()
            return ':f:func:`%(fname)s`'%locals()
        
        # Remote reference
        from fortran_domain import f_sep
        if falias:
            ':f:func:`%(falias)s<~%(module)s%(f_sep)s%(fname)s>`'%locals()
        return ':f:func:`~%(module)s%(f_sep)s%(fname)s`'%locals()
        

            
    def format_use(self, block, indent=0, short=False):
        """Format use statement
        
        :Parameters:
        
            - *block*: a module block
        """
        # TODO: format aliases bug
        use = ''
        if block.has_key('use'):
            if short:
                use = ':use: '
            else:
                use = self.format_subsection('Needed modules', indent=indent)
            lines = []
            for mname, monly in block['use'].items():
                
                # Reference to the module
                line = (self.indent(indent) if not short else '')+':f:mod:`%s`'%mname
                
                # Reference to the routines
                if monly:
                    funcs = []
                    for fname, falias in monly['map'].items():
                        func = self.format_funcref(fname, module=mname)
                        if fname!=falias:
                            falias = self.format_funcref(falias, module=mname, aliasof=fname)
                            func = '%s => %s' % (falias, func)
                        funcs.append(func)
                    line += ' (%s)'% ', '.join(funcs)
                    
                # Short description
                if self.modules.has_key(mname) and not short:
                    sdesc = self.get_synopsis(self.modules[mname])
                    if sdesc:
                        line += ': '+sdesc
                    
                # Append
                lines.append(line)
            if short:
                use += ', '.join(lines)
                use = self.format_lines(use,indent)
            else:
                use += self.format_lines(lines, indent, bullet='-') + '\n'
            del lines
        return use
            
    #def format_arithm(self, expr):
        #"""Format an arithmetic expression"""
        #ops = re.findall(r'(\W+)',expr)
        #nums = re.split(r'\W+', expr)
        #if expr.startswith(ops[0]): 
            #nums = ['']+nums
        #if len(nums)!=len(ops): ops.append('')
        #newexpr = ''
        #for num, op in zip(nums, ops):
            #if num.isalpha():
                #num = ' :f:var:`%s` '%num
            #if '*' in op:
                #op = op.replace('*', '\*')
            #newexpr += num+op
        #return newexpr              
               
    def format_argdim(self, block):
        """Format the dimension of a variable
        
         :Parameters:
        
            - *block*: a variable block
        """
        if block.has_key('dimension'):
            return'(%s)'%(','.join([s.strip('()') for s in block['dimension']]))
            #return'(%s)'%(','.join([s.replace(':','\:').strip('()') for s in block['dimension']]))
        return ''

    def format_argattr(self, block):
        """Filter and format the attributes (optional, in/out/inout, etc) of a variable
        
         :Parameters:
        
            - *block*: a variable block
        """
        vattr = []
        if block.has_key('intent') and block['intent']: 
            vattr.append('/'.join(block['intent']))
        if block.has_key('attrspec') and block['attrspec']: 
            newattrs = []
            for attr in block['attrspec']:
                if '=' in block:
                    if attr=='optional':
                        continue
                    elif attr=='parameter':
                        attr += '='+block['=']
                        #attr += '='+self.format_arithm(block['='])
                if attr in []:
                    attr = attr.upper()
                newattrs.append(attr)
            block['attrspec'] = newattrs
            vattr.append('/'.join(block['attrspec']))
        if not vattr: return ''
        vattr = ','.join(vattr)
        return self._fmt_vattr%locals() if vattr else ''
        
    def format_argtype(self, block):
        vtype = block['typespec']
        if vtype=='type': vtype = block['typename']
        return vtype
 
    def format_argfield(self, blockvar, role=None, block=None):
        """Format the description of a variable
        
         :Parameters:
        
            - *block*: a variable block
        """
        vname = blockvar['name']
        vtype = self.format_argtype(blockvar)#['typespec']
        #if vtype=='type': vtype = block['typename']
        
        vdim = self.format_argdim(blockvar)
        vattr = self.format_argattr(blockvar)
        vdesc = blockvar['desc'] if blockvar.has_key('desc') else ''
        optional = blockvar.has_key('attrspec') and 'optional' in blockvar['attrspec']
        if not role:
            if block and vname in [block['name'], block.get('result')]:
                role = 'r'
            else:
                role = 'o' if optional else 'p'
        return self._fmt_vardesc%locals()
       
    def format_type(self, block, indent=0, bullet=True):
        """Format the description of a module type
        
         :Parameters:
        
            - *block*: block of the type
        """
        # Declaration and description
        declaration = self.format_declaration('type', block['name'], block['desc'], indent=indent, bullet=bullet)+'\n'
        
        # Variables
        vlines = []
        for bvar in block['vars'].values():
            vlines.append(self.format_argfield(bvar, role='f'))
        variables = self.format_lines(vlines, indent=indent+1)+'\n'
        del vlines
        
        return declaration+variables 
    
        
    def get_varopts(self, block):
        """Get options for variable declaration as a dict"""
        options = {}
        vdim = self.format_argdim(block)
        #if vdim!='': vdim = self._fmt_fvardim%locals()
        if vdim:
            options['shape'] = vdim
        options['type'] = self.format_argtype(block)
        vattr = self.format_argattr(block).strip(' []')
        if vattr:
            options['attrs'] = vattr
        return options
       
    #def format_vardesc(self, block):
        #"""Format the specification of a variable at top of its declaration content"""
        #specs = self.format_varspecs(block)
        #vdesc = block.get('desc', '')
        #return specs+'\n'+vdesc+'\n'
        
    def format_var(self, block, indent=0, bullet=True):
        """Format the description of a module type
        
         :Parameters:
        
            - *block*: block of the variable
        """
        # Description of the variable
        options = self.get_varopts(block)
        description = block.get('desc', None)
        declaration = self.format_declaration('variable', block['name'], description=description, 
            options=options, indent=indent, bullet=bullet)

        ## Description of the sub-variables
        #if block['typespec']=='type':
            #xxxx
            #variables = []
            #btype = self.types[block['typename']]
            #for bvar in btype['vars'].values():
                #vname = '%s%%%s'%(block['name'], bvar['name'])
                #desc = self.format_vardesc(bvar, indent=0, vname=vname)
                #variables.append(self.format_declaration('var', vname, desc, indent=indent+1, bullet=True)[:-1])
            #variables = '\n'.join(variables)
        #else:
            #variables = ''

        return declaration#+variables
       
    
    def format_signature(self, block):
        signature = ''
        nopt = 0
        for i, var in enumerate(block['args']):
            optional  = 'optional' in block['vars'][var]['attrspec'] and '=' not in block['vars'][var] \
                if block['vars'][var].has_key('attrspec') else False
            signature += '[' if optional  else ''
            signature += ', ' if i else ''
            if optional: nopt +=1
            signature += var
        signature += nopt*']'
        return signature
       
    
    def format_routine(self, block, indent=0):
        """Format the description of a function, a subroutine or a program"""
        # Declaration of a subroutine or function
        if isinstance(block, basestring): 
            if block not in self.programs.keys()+self.routines.keys():
                raise F90toRstException('Unknown function, subroutine or program: %s'%block)
            if block in self.programs:
                block = self.programs[block]
            else:
                block = self.routines[block]
        elif block['name'] not in self.modules.keys()+self.routines.keys():
                raise F90toRstException('Unknown %s: %s'%(block['type'], block['name']))

        name = block['name']
        blocktype = block['block']
        signature = '(%s)'%self.format_signature(block) if blocktype!='program' else ''
        declaration = self.format_declaration(blocktype, 
            '%(name)s%(signature)s'%locals(), indent=indent)
        #declaration = self.indent(indent)+'.. f:%(blocktype)s:: %(name)s%(signature)s\n\n'%locals()
        
        # Treat variables in comment (subroutines and functions only)
        comments = list(block['desc'])+['']
        if blocktype!='program' :
            found = []
            for iline in xrange(len(comments)):
                m = block['vardescmatch'](comments[iline])
                if m:
                    varname = m.group('varname')
                    found.append(varname)
                    comments[iline] = self.format_argfield(block['vars'][varname], block=block)                
            for varname in block['args']+block['sortvars']:
                if varname not in found:
                    comments.append(self.format_argfield(block['vars'][varname], block=block))
                    found.append(varname)
        
        # Description
        description = self.format_lines(comments, indent+1)
        
        # Add use of modules
        use = self.format_use(block, indent=indent+1, short=True)
        
        # Add calls
        calls = []
        module  = block.get('module')
        # - call froms
        if blocktype in ['function', 'subroutine']:
            if block['callfrom']:
                callfrom = []
               
                for fromname in block['callfrom']:
                    if fromname in self.routines:
                        cf = self.format_funcref(fromname, module)
                    else:
                        cf = ':f:prog:`%s`'%fromname
                    callfrom.append(cf)
                    
                #callfrom += ', '.join([self.format_funcref(getattr(self, routines[fn]['name'], module) for fn in block['callfrom']])
                callfrom = ':from: ' + ', '.join(callfrom)

                calls.append(callfrom)
        # - call tos
        if block['callto']:
            callto = ', '.join([self.format_funcref(fn, module) for fn in block['callto']])
            #callto = ', '.join([self.format_funcref(self.routines[fn]['name'], module) for fn in block['callto']])
            if callto == '': callto = 'None'
            callto = ':to: '+callto
            calls.append(callto)
        calls = '\n'+self.format_lines(calls, indent=indent+1)
        return declaration+description+use+calls+'\n\n'
        
    format_function = format_routine
    format_subroutine = format_routine 
        
        
    def format_quickaccess(self, module, indent=indent):
        """Format an abstract of all types, variables and routines of a module"""
        if not isinstance(module, basestring): module = module['name']
        
        # Title
        title = self.format_subsection('Quick access', indent=indent)+'\n'
        
        # Types
        decs = []
        tlist = self.get_blocklist('types', module)
        tlist.sort()
        if tlist:
            decs.append(':Types: '+', '.join([':f:type:`%s`'%tt['name'] for tt in tlist]))
        
        # Variables
        vlist = self.get_blocklist('variables', module)
        vlist.sort()
        if vlist:
            decs.append(':Variables: '+', '.join([':f:var:`%s`'%vv['name'] for vv in vlist]))
        
        # Functions and subroutines
        flist = self.get_blocklist('functions', module)
        flist.sort()
        if flist:
            decs.append(':Routines: '+', '.join([':f:func:`~%s/%s`'%(module, ff['name']) for ff in flist]))
        
        if decs: return self.format_lines(title+'\n'.join(decs))+'\n\n'
        return ''
        
    def format_types(self, block, indent=0):
        """Format the description of all fortran types"""
        types = []
        for subblock in block['body']:
            if subblock['block']=='type':
                types.append(self.format_type(subblock, indent=indent))
        if types: 
            types = self.format_subsection('Types', indent=indent)+'\n'.join(types)
        else:
            types = ''
        return types
        
    def format_variables(self, block, indent=0):
        """Format the description of all variables (global or module)"""
        variables = ''
        if block['vars']:
            for bvar in block['vars'].values():
                variables += self.format_var(bvar, indent=indent)
            variables = self.format_subsection('Variables', indent=indent)+variables+'\n\n'
        return variables
        
    def format_description(self, block, indent=0):
        """Format the description of an object"""
        description = ''
        if block['desc']:
            description = self.format_subsection('Description', indent=indent)
            description += self.format_lines(block['desc'],indent=indent, strip=True)+'\n'
        return description
        
    def format_routines(self, block, indent=0):
        """Format the list of all subroutines and functions"""
        routines = ''
        if block['body']:
            fdecs = []
            for subblock in block['body']:
                if subblock['block'] in ['function', 'subroutine']:
                    fdecs.append(self.format_routine(subblock, indent))
            if fdecs:
                fdecs = '\n'.join(fdecs)
                routines = self.format_subsection('Subroutines and functions', indent=indent)+fdecs
        return routines
        
    def format_module(self, block, indent=0):
        """Recursively format a module and its declarations"""

        # Declaration of the module
        if isinstance(block, basestring): 
            if block not in self.modules:
                raise F90toRstException('Unknown module: %s'%block)
            block = self.modules[block]
        elif block['name'] not in self.modules:
                raise F90toRstException('Unknown module: %'%block['name'])

        modname = block['name']
        declaration = self.format_declaration('module', modname, indent=indent,
           options=dict(synopsis = self.get_synopsis(block).strip() or None))

        # Description
        description = self.format_description(block, indent=indent)

        # Quick access
        quickaccess = self.format_quickaccess(modname, indent=indent)

        # Use of other modules
        use = self.format_use(block, indent=indent)

        # Types
        types = self.format_types(block, indent=indent)
                 
        # Variables
        variables = self.format_variables(block, indent=indent)
                
        # Subroutines and functions
        routines = self.format_routines(block, indent=indent)                
               
        return declaration + description + quickaccess + use + types + variables + routines
            
    def format_srcfile(self, srcfile, indent=0, **kwargs):
        """Format all declaration of a file, except modules"""
        rst = ''
        # Programs
        bprog = self.filter_by_srcfile(srcfile, objtype='program', **kwargs)
        if bprog:
            programs = self.format_subsection('Program', indent=indent)+'\n'
            rst += self.format_program(bprog[0], indent=indent)+'\n'
            
        # Functions and subroutines
        brouts = self.filter_by_srcfile(srcfile, objtype=['function', 'subroutine'], **kwargs)
        if brouts:
            self.format_subsection('Functions et subroutines', indent=indent)+'\n'
            for block in brouts:
                rst += self.format_routines(block, indent=indent)+'\n'
                
        return rst
        
    def __getitem__(self, module):
        return self.format_module(self.modules[module])
        

# Sphinx directive ---

def setup(app):
    
    app.add_description_unit('ftype', 'ftype', indextemplate='pair: %s; Fortran type', )
    app.add_description_unit('fvar', 'fvar', indextemplate='pair: %s; Fortran variable', )
    
    app.add_config_value('fortran_title_underline', '-', False)
    app.add_config_value('fortran_indent', 4, False)
    app.add_config_value('fortran_subsection_type', 'rubric', False)
    app.add_config_value('fortran_src', '.', False)
    app.add_config_value('fortran_ext', ['f90', 'f95'], False)
    app.add_config_value('fortran_encoding', 'utf8', False)
    
    #app.add_directive('fortran_module', IncludeFortranDirective)
    FortranDomain.directives.update(
        automodule=FortranAutoModuleDirective,
        autoroutine=FortranAutoObjectDirective,
        autofunction=FortranAutoFunctionDirective,
        autosubroutine=FortranAutoSubroutineDirective,
        autoprogram=FortranAutoProgramDirective,
        autosrcfile=FortranAutoSrcfileDirective,
    )
    app.connect('builder-inited', fortran_parse)

def list_files(fortran_src, exts=['f', 'f90', 'f95'], absolute=True):
    """Get the list of fortran files"""

    # Extensions (suffixes)
    if not isinstance(exts, list):
        exts = list(exts)
    for e in exts:
        if e.lower() not in exts: exts.append(e.lower())
        if e.upper() not in exts: exts.append(e.upper())
        
    # List the files using globs
    ffiles = []
    for fg in fortran_src:
        if not isinstance(fg, basestring): continue
        if os.path.isdir(fg):
            for ext in exts:
                ffiles.extend(glob(os.path.join(fg,'*.'+ext)))
        else:
            ffiles.extend(glob(fg))
    if absolute:
        ffiles = [os.path.abspath(ffile) for ffile in ffiles]
    return ffiles

def fortran_parse(app):
    env = app.builder.env
    if isinstance(app.config.fortran_src, (str, list)):
        app.info('Parsing fortran sources...', nonl=True)
        
        # Sources a list
        if not isinstance(app.config.fortran_src, list):
            app.config.fortran_src = [app.config.fortran_src]
        
        # All files
        ffiles = list_files(app.config.fortran_src, app.config.fortran_ext)

        # Parse files
        if not ffiles:
            app.warn("No fortran files found")
            app.config._f90torst = None
            app._status.write('\n')
        else:
            app.config.fortran_indent = fmt_indent(app.config.fortran_indent)
            app.config._f90torst = F90toRst(ffiles, ic=app.config.fortran_indent, 
                ulc=app.config.fortran_title_underline, encoding=app.config.fortran_encoding)
            app._status.write('done\n')
        app._status.flush()
        
    else:
        app.warn("Wrong list of fortran 90 source specifications: "+str(app.config.fortran_src))
        app.config._f90torst = None
#    app.config._f90files = []


       
def fmt_indent(string):    
    if string is None: return
    if isinstance(string, int): string = ' '*string
    if string == 'tab': string = '\t'
    elif string == 'space': string = ' '
    return string


#class fortran_module(nodes.General, nodes.Element):
    #pass


class FortranAutoModuleDirective(Directive):
    has_content = True
    option_spec = dict(title_underline=unchanged, indent=fmt_indent, 
        subsection_type=unchanged)
    required_arguments = 1
    optional_arguments = 0
    
    def run(self):

        # Get environment
        f90torst = self.state.document.settings.env.config._f90torst
        if f90torst is None: return []
        

        # Check module name
        module = self.arguments[0]
        if not f90torst.modules.has_key(module):
            self.warn('Wrong fortran module name: '+module)
        
        # Options
        ic = f90torst.ic
        ulc = f90torst.ulc
        if self.options.get('indent'):
            f90torst.ic = self.options['indent']
        if self.options.get('title_underline'):
            f90torst.ulc = self.options['title_underline']
        if self.options.get('subsection_type'):
            f90torst.ulc = self.options['subsection_type']
        
        # Get rst
        raw_text = f90torst.format_module(module)
        
        # Insert it
        source = self.state_machine.input_lines.source(
            self.lineno - self.state_machine.input_offset - 1)
        include_lines = string2lines(raw_text, convert_whitespace=1)
        self.state_machine.insert_input(include_lines,source)
    
        # Restore defaults
        if self.options.has_key('indent'): f90torst.ic = ic
        if self.options.has_key('title_underline'): f90torst.ulc = ulc
        if self.options.has_key('subsection_type'): f90torst.sst = sst
    
        return []

 
class FortranAutoObjectDirective(Directive):
    """Generic directive for fortran object auto-documentation
    
    Redefine :attr:`_warning` and :attr:`_objtype` attribute when subcassling.
    
    .. attribute:: _warning
    
        Warning message when object is not found, like:
        
        >>> _warning = 'Wrong function or subroutine name: %s'
        
    .. attribute:: _objtype
    
        Type of fortran object.
        If "toto" is set as object type, then :class:`F90toRst` must have 
        attribute :attr:`totos` containg index of all related fortran objects,
        and method :meth:`format_totos` for formatting the object.
            
    """
    has_content = False
    option_spec = {}
    required_arguments = 1
    optional_arguments = 0
    _warning = 'Wrong routine name: %s'
    _objtype = 'routine'
    
    def run(self):

        # Get environment
        f90torst = self.state.document.settings.env.config._f90torst
        if f90torst is None: return []

        # Check object name
        objname = self.arguments[0].lower()
        from fortran_domain import f_sep
        if f_sep in objname: objname = objname.split(f_sep)[-1] # remove module name
        objects = getattr(f90torst, self._objtype+'s')
        if not objects.has_key(objname):
            self.warn(self._warning%objname)
                
        # Get rst
        raw_text = getattr(f90torst, 'format_'+self._objtype)(objname)
        
        # Check if inside module
        b = objects[objname]
        if b.has_key('parent_block'):
            curmod_text = '.. f:currentmodule:: %s\n\n'%b['parent_block']['name']
            raw_text = curmod_text+raw_text

        # Insert it
        source = self.state_machine.input_lines.source(
            self.lineno - self.state_machine.input_offset - 1)
        include_lines = string2lines(raw_text, convert_whitespace=1)
        self.state_machine.insert_input(include_lines,source)
        
        return []
 
        
class FortranAutoFunctionDirective(FortranAutoObjectDirective):
    _warning = 'Wrong function name: %s'
    _objtype = 'function'
    

class FortranAutoSubroutineDirective(FortranAutoObjectDirective):
    _warning = 'Wrong subroutine name: %s'
    _objtype = 'subroutine'
    

class FortranAutoTypeDirective(FortranAutoObjectDirective):
    _warning = 'Wrong type name: %s'
    _objtype = 'type'
    
class FortranAutoVariableDirective(FortranAutoObjectDirective):
    _warning = 'Wrong variable name: %s'
    _objtype = 'variable'

class FortranAutoProgramDirective(Directive):
    has_content = False
    option_spec = {}
    required_arguments = 1
    optional_arguments = 0
    
    def run(self):

        # Get environment
        f90torst = self.state.document.settings.env.config._f90torst
        if f90torst is None: return []

        # Check routine name
        program = self.arguments[0].lower()
        if not f90torst.programs.has_key(program):
            self.warning('Wrong program name: '+program)
                
        # Get rst
        raw_text = f90torst.format_routine(program)
        
        # Insert it
        source = self.state_machine.input_lines.source(
            self.lineno - self.state_machine.input_offset - 1)
        include_lines = string2lines(raw_text, convert_whitespace=1)
        self.state_machine.insert_input(include_lines,source)
        
        return []

class FortranAutoSrcfileDirective(Directive):
    has_content = False
    option_spec = dict(search_mode=unchanged, objtype=unchanged)
    required_arguments = 1
    optional_arguments = 0
    
    def run(self):

        # Get environment
        f90torst = self.state.document.settings.env.config._f90torst
        if f90torst is None: return []

        # Check routine name
        srcfile = self.arguments[0].lower()
        if not f90torst.programs.has_key(program):
            self.warning('Wrong program name: '+program)
                
        # Options
        search_mode = self.options.get('search_mode')
        objtype = self.options.get('objtype')
        if objtype:
            objtype = objtype.split(' ,')
        
        # Get rst
        raw_text = f90torst.format_srcfile(srcfile, search_mode=search_mode, objtype=objtype)
        if not raw_text:
            self.warning('No valid content found for file: '+srcfile)
        
        # Insert it
        source = self.state_machine.input_lines.source(
            self.lineno - self.state_machine.input_offset - 1)
        include_lines = string2lines(raw_text, convert_whitespace=1)
        self.state_machine.insert_input(include_lines,source)
        
        return []




#! /usr/bin/env python
#===============================================================================
# options.preferences_manager
#===============================================================================
# Copyright (C) 2012-2013 Christian Fernandez
#       Laboratoire Catalyse et Spectrochimie, Caen, France.  
#       christian.fernandez@ensicaen.fr
# This software is governed by the CeCILL-B license under French law 
# and abiding by the rules of distribution of free software.  
# You can  use, modify and/ or redistribute the software under 
# the terms of the CeCILL-B license as circulated by CEA, CNRS and INRIA
# at the following URL "http://www.cecill.info".
# See Licence.txt in the main masai source directory
#===============================================================================

""" Masai preferences manager. """

#===============================================================================
# Enthought library imports.
#===============================================================================
from traits.api import Color, Str, Float, Bool
from traitsui.api import View, Item, Group, VGroup, spring
import os

#===============================================================================
# Local imports.
#===============================================================================
from apptools.preferences.api import Preferences
from apptools.preferences.api import get_default_preferences
from apptools.preferences.api import set_default_preferences
from apptools.preferences.ui.api import PreferencesManager, PreferencesPage

__all__ = ['get_categories', 'get_options_names', 
           'get_options', 'set_options', 'edit_options']

# Create a preferences collection from a file and make it the default root
# preferences node for all preferences helpers etc.
# we don't need this when running the program from main.py
filename = os.path.join("~",".masai","masai.ini")
set_default_preferences(Preferences(filename=os.path.expanduser(filename)))

#===============================================================================
# MasaiPreferencesPage
#===============================================================================
class MasaiAppPreferencesPage(PreferencesPage):
    """ A preference page for the Masai preferences. """

    #### 'IPreferencesPage' interface #########################################

    # The page's category (e.g. 'General/Appearence'). The empty string means
    # that this is a top-level page.
    category = ''

    # The page's help identifier (optional). If a help Id *is* provided then
    # there will be a 'Help' button shown on the preference page.
    help_id = ''

    # The page name (this is what is shown in the preferences dialog.
    name = 'Application'

    # The path to the preferences node that contains our preferences.
    preferences_path = 'application'
    
    #### Preferences ##########################################################

    last_source_file = Str('')
    path_of_data = Str( os.path.expanduser('~/bruker/Data'))
    width = Float(1.0)
    height = Float(0.9)
     
    #### Traits UI views ######################################################

    view = View(Group(
                    VGroup(
                           Item('last_source_file'), 
                           Item('path_of_data'),
                           label = 'Source paths',
                          ),
                    VGroup( 
                           Item('width'),
                           Item('height'),
                           spring,
                           label = 'main window',
                           ),
                    orientation = 'vertical',
                    ),
                )

#===============================================================================
# MasaiProcPreferencesPage
#===============================================================================
class MasaiProcPreferencesPage(PreferencesPage):
    """ A preference page for the Masai preferences. """

    #### 'IPreferencesPage' interface #########################################

    # The page's category (e.g. 'General/Appearence'). The empty string means
    # that this is a top-level page.
    category = ''

    # The page's help identifier (optional). If a help Id *is* provided then
    # there will be a 'Help' button shown on the preference page.
    help_id = ''

    # The page name (this is what is shown in the preferences dialog.
    name = 'Processing'

    # The path to the preferences node that contains our preferences.
    preferences_path = 'processing'
    
    #### Preferences ##########################################################

    process_automatically = Bool(True)
    
     
    #### Traits UI views ######################################################

    view = View('''process_automatically'''
                )

#===============================================================================
# MasaiFilesPreferencesPage
#===============================================================================
class MasaiPlotPreferencesPage(PreferencesPage):
    """ A preference page for the Masai workbench preferences. """

    #### 'IPreferencesPage' interface #########################################

    # The page's category (e.g. 'General/Appearence'). The empty string means
    # that this is a top-level page.
    category = ''

    # The page's help identifier (optional). If a help Id *is* provided then
    # there will be a 'Help' button shown on the preference page.
    help_id = ''

    # The page name (this is what is shown in the preferences dialog.
    name = 'Plot'

    # The path to the preferences node that contains our preferences.
    preferences_path = 'plot'

    #### Preferences ##########################################################

    bgcolor = Color('white')
    fgcolor = Color('black')

    #### Traits UI views ######################################################

    view = View('bgcolor', 'fgcolor')

def get_categories():
    """Return a list of option categories for the application
    
    Examples
    --------
    >>> get_categories()
    ['application', 'processing', 'root', 'plot']
    
    """
    return get_default_preferences().node_names()

def get_options_names(category=None):
    """Get list of option names for a given category.
    
    Parameters
    ----------
    category : string
        The category to look for.
        
    Examples
    --------

    >>> from masai.api import get_options
    >>> get_options_names('application')    
    ['last_source_file', 'path_of_data', 'width', 'height']
    
    """
    if not category:
        return get_default_preferences().dump()
    
    return get_default_preferences().node(category).keys()

def get_options(category, option):
    """Get the value of a given option
    
    Parameters
    ----------
    category : string
        The category of preference
    option : string
        The option to get
        
    Examples
    --------

    >>> from masai.api import set_options
    >>> get_options('application','path_of_data') 
    '/users/christian/bruker/Data'
    
    """
    
    pref = get_default_preferences()
    return pref.get("%s.%s"%(category,option))
    
    

def set_options(category, option, value):
    """Set a given option to some value
    
    Parameters
    ----------
    category : string
        The category of preference
    option : string
        The option to set
    value : any
        The value of the option to set
        
    Examples
    --------

    >>> from masai.api import set_options
    >>> set_options('application','path_of_data','/users/christian/bruker/Data')    
    
    """
    
    pref = get_default_preferences()
    pref.set("%s.%s"%(category,option),value)
    pref.flush()
    
def edit_options():
    """Open a view to manage the options/preference for the whole application
    
    Examples
    --------
   
    >>> from masai.api import edit_options
    >>> edit_options()

    """
    option_manager = PreferencesManager(
        pages = [
            MasaiAppPreferencesPage(), 
            MasaiProcPreferencesPage(), 
            MasaiPlotPreferencesPage()
        ]
    )

    # Show the UI...
    option_manager.configure_traits()

    # Save the preferences...
    get_default_preferences().flush()
    
# Entry point.
if __name__ == '__main__':

    import nose
    nose.main()

    
#### EOF ######################################################################



#! /usr/bin/env python
#===============================================================================
# controllers.proccontroller
#===============================================================================
# Copyright (C) 2012-2013 Christian Fernandez
#       Laboratoire Catalyse et Spectrochimie, Caen, France.  
#       christian.fernandez@ensicaen.fr
# This software is governed by the CeCILL-B license under French law 
# and abiding by the rules of distribution of free software.  
# You can  use, modify and/ or redistribute the software under 
# the terms of the CeCILL-B license as circulated by CEA, CNRS and INRIA
# at the following URL "http://www.cecill.info".
# See Licence.txt in the main masai source directory
#===============================================================================
"""
.. _proccontroller:

This module defines objets and methods related to the processing of the
datasets

"""
#===============================================================================
# Enthought imports
#===============================================================================
from traits.api import (HasTraits, Instance, Range, Bool, Int, Enum,
                        Float, Button, on_trait_change)
from traitsui.api import (View, Item, Group, VGroup, HGroup,
                          spring, ScrubberEditor)
     
#===============================================================================
# numpy
#===============================================================================
import numpy as np

#===============================================================================
# logger
#===============================================================================
import logging
logger = logging.getLogger()

#===============================================================================
# masai imports
#===============================================================================
from masai.api import (tdeff, em, pk, apk, efp, largest_power_of_2)

#===============================================================================
# formatter
#===============================================================================
def formatter(value):
    """Format of an Item
    """
    if isinstance(value, float):
        return str("%.3f" % value)
    else:
        return str(value)
     
#===============================================================================
# ProcController
#===============================================================================
class ProcController(HasTraits):
    """
    This controller hosts the method for processing the data displyed in the 
    viewplot..
    
    """
    # the figure where to plot our data
    #viewplot = Instance(ViewPlot)

    # the source of data
    fd = Instance('masai.fileio.source.Source')
    
    #===========================================================================
    # # other traits for processing    
    #===========================================================================
    autoprocessing = Bool(False)
    td = Int
    
    tdeff = Range(low=256)
    def _tdeff_changed(self):
        self.fd.sourcepref.set('processing.tdeff', self.tdeff)
        self.fd.sourcepref.flush()
        self.fd.TDeff[-1] = self.tdeff
        tdeff(self.fd)
        
    si = Enum(2 ** 10, 2 ** 11, 2 ** 12, 2 ** 13, 2 ** 14, 2 ** 15, 2 ** 16)
    def _si_changed(self):
        self.fd.sourcepref.set('processing.si', self.si)
        self.fd.sourcepref.flush()
        self.fd.SI[-1] = self.si
    
    lb = Range(low=0.)
    def _lb_changed(self):
        self.fd.sourcepref.set('processing.lb', self.lb)
        self.fd.sourcepref.flush()
        self.fd.LB[-1]=self.lb
        em(self.fd)

    ph0 = Range(0., +360.)
    ph1 = Float
    ph0inc = 10.
    ph1inc = 10.

    def _ph0_changed(self):
        if not self.fd.is_freq[-1]: return
        self.fd.PHC0[-1] = self.ph0
        self.fd.sourcepref.set('processing.ph0', self.ph0)
        self.fd.sourcepref.flush()
        pk(self.fd)
        
    def _ph1_changed(self):
        if not self.fd.is_freq[-1]: return 
        self.fd.PHC1[-1] = self.ph1
        self.fd.sourcepref.set('processing.ph1', self.ph1) 
        si = self.fd.data.shape[-1]
        pcor = np.deg2rad(self.ph1 - self.fd.PHC1SAV[-1])*\
                                             float(self.fd.phcursor) / float(si)
        self.ph0 = (self.ph0 - np.rad2deg(pcor)) % 360.
        
    apk = Button("APK")
    def _apk_fired(self):
        apk(self.fd)
        self.ph0 = self.fd.PHC0[-1] % 360.0 
        
    inv = Button(u'ADD 180\u00B0')
    def _inv_fired(self):
        self.ph0 = (self.fd.PHC0[-1] - 180.) % 360.

    quad = Button(u'ADD 90\u00B0')
    def _quad_fired(self):
        self.ph0 = (self.fd.PHC0[-1] + 90.) % 360.
                              
    efp = Button("EFP")
    def _efp_fired(self):
        self.fd.can_plot = False
        efp(self.fd) 
        self.fd.sourcepref.set('processing.forcefid', False)
        self.fd.sourcepref.flush()
        self.fd.can_plot = True
                                                                 
    reset = Button("RESET TO ORIGINAL")    
    def _reset_fired(self):
        self.fd.reset()
        self.resetpar()

#    baseline = Instance(Baseline)
#    def _baseline_default(self):
#        return Baseline(fd=self.fd)
#    
    #===========================================================================
    # view
    #===========================================================================
    def default_traits_view(self):
        return View(
              Group(
                    VGroup(HGroup(
                           Item('ph0', width=100,
                                   editor=ScrubberEditor(
                                                        format_func=formatter,
                                                        increment=.5,
                                                        hover_color=0xFFFFFF,
                                                        active_color=0xB0DD8E),
                                ),
                           Item('ph1', width=100,
                                   editor=ScrubberEditor(
                                                        format_func=formatter,
                                                        increment=1.0,
                                                        hover_color=0xFFFFFF,
                                                        active_color=0xB0DD8E)

                            ),
                           ),
                           HGroup(
                                  spring,
                                  Item('apk', show_label=False),
                                  Item('quad', show_label=False),
                                  Item('inv', show_label=False),
                           ),
                           Item('_'),
                           enabled_when='fd.is_freq[-1]',
                           label='Phases',
                    ),
                    VGroup(
                        HGroup(
                           Item('lb',
                                editor=ScrubberEditor(
                                                      format_str="%.2f",
                                                      increment=.5,
                                                      hover_color=0xFFFFFF,
                                                      active_color=0xB0DD8E)
                                ),
                           #Item('reducesampling', label='Reduce sampling'),
                           Item('tdeff', label='effective TD',
                                 editor=ScrubberEditor(
                                                       format_str="%d",
                                                       increment=1,
                                                       hover_color=0xFFFFFF,
                                                       active_color=0xB0DD8E)),
                           spring,
                        ),
                       Item('_'),
                       label='Apodization',
                       enabled_when='not fd.is_freq[-1]',
                    ),
                    VGroup(
                        HGroup(
                           Item('si', label='transformed size'),
                           Item('efp', show_label=False),
                           spring,
                        ),
                       Item('_'),
                       enabled_when='not fd.is_freq[-1]',
                       label='Transforms',
                    ),
                  Item('reset', show_label=False),
                  visible_when='fd'
               ),
            resizable=True,
            width=0.33, #TODO: find the way to set the initial width
            height=1.,
#            handler = ProcControlHandler()
            )
        
    def resetpar(self):
        """Reset some parmeters to their default value (
           (or the preferred value in the preference file whne it exists)
           
        """
        self.tdeff = self.fd.TDeff[-1] = int(self.fd.sourcepref.get(
                   'processing.tdeff', max(self.fd.TD[-1], self.fd.TDeff[-1])))
        self.si = self.fd.SI[-1] = int(self.fd.sourcepref.get('processing.si', 
                                        largest_power_of_2(self.fd.TD[-1] * 2)))
        self.lb = self.fd.LB[-1] = float(self.fd.sourcepref.get('processing.lb',
                                                                self.fd.LB[-1]))
        self.ph0 = self.fd.PHC0[-1] = float(self.fd.sourcepref.get(
                                    'processing.ph0', self.fd.PHC0[-1])) % 360.0
        self.ph1 = self.fd.PHC1[-1] = float(self.fd.sourcepref.get(
                                            'processing.ph1', self.fd.PHC1[-1]))
        
    #===========================================================================
    # event router
    #===========================================================================
    @on_trait_change('fd.+')
    def _eventrouter(self, objet, name, old, new):
        
        if not self.fd.is_loaded: 
            # we return if the source is not valid
            return   
         
        if name == "fd":
            # if we have this event, 
            # this is because the fd source has just been updated
            # get previous values from preference files
            logger.debug("proccontroller: %s from %s to %s" % (name, str(old), 
                                                               str(new)))

            self.resetpar()        



#!/usr/bin/python
# -*- coding: utf-8 -*-
#===============================================================================
# __init__
#===============================================================================
# Copyright (C) 2012-2013 Christian Fernandez
#       Laboratoire Catalyse et Spectrochimie, Caen, France.  
#       christian.fernandez@ensicaen.fr
# This software is governed by the CeCILL-B license under French law 
# and abiding by the rules of distribution of free software.  
# You can  use, modify and/ or redistribute the software under 
# the terms of the CeCILL-B license as circulated by CEA, CNRS and INRIA
# at the following URL "http://www.cecill.info".
# See Licence.txt in the main masai source directory
#===============================================================================

"""The **MASAÏ** package is intended to provide two main services:  

1. an API, which can be used in other user applications

2. a GUI, providing extensible tools for processing, analysing and fitting of 
   solid state NMR spectra. 

.. note:: 

    In version 0.1, only 1D processing and fitting tools are provided

.. seealso:: the :ref:`installation` guide for install instructions


"""


#! /usr/bin/env python
#===============================================================================
# main
#===============================================================================
# Copyright (C) 2012-2013 Christian Fernandez
#       Laboratoire Catalyse et Spectrochimie, Caen, France.  
#       christian.fernandez@ensicaen.fr
# This software is governed by the CeCILL-B license under French law 
# and abiding by the rules of distribution of free software.  
# You can  use, modify and/ or redistribute the software under 
# the terms of the CeCILL-B license as circulated by CEA, CNRS and INRIA
# at the following URL "http://www.cecill.info".
# See Licence.txt in the main masai source directory
#===============================================================================
"""
.. _controller:

This module defines objets and methods related to the control of the dataset model.
It is displayed in the left side of main application window

"""
#===============================================================================
# python import
#===============================================================================
import logging
import time

#===============================================================================
# Enthought imports
#===============================================================================
from traits.api import HasTraits, Instance, on_trait_change
from traitsui.api import View, Group

#===============================================================================
# masai imports
#===============================================================================
from masai.utils import InstanceUItem
from fitcontroller import FitController
from plotcontroller import PlotController
from proccontroller import ProcController

#===============================================================================
# logger
#===============================================================================
logger = logging.getLogger()

#===============================================================================
# Controllers
#===============================================================================
class Controllers(HasTraits):
    """Define a sub view for handling the controllers on the left side 
    of the main window

    Three controllers with a GUI displayed in a separate notebook tab are available:
    
        * ProcController, to handle everything related to processing of the spectra
        
        .. seealso:: :ref:`procontroller module <proccontroller>`
           
        * PlotController to manage the display of the data
        
        .. seealso:: :ref:`plotcontroller module <plotcontroller>`
        
        * FitController to set and perform fit of the spectra
        
        .. seealso:: :ref:`fitcontroller module <fitcontroller>`
        
    .. todo::
    
        Add analysis capabilities (integration, peak peaking, and so on...)
    
    """
    #: Instance of class :class:`Source` 
    fd = Instance('masai.api.Source')

    #: Instance of class :class:`ProcController` 
    proccontroller = Instance(ProcController)
    
    #: Instance of class :class:`FitController` 
    fitcontroller = Instance(FitController)
    
    #: Instance of class :class:`PlotController` 
    plotcontroller = Instance(PlotController)
    
    #: Instance of class :class:`ViewPlot` 
    viewplot = Instance('masai.views.ViewPlot')
    
    def default_traits_view(self):
        """Return the main controller view
        
        This view displays a notebook on the left side of the main view 
        with severall controllers (proc, fit and plotcontroller), 
        in a separate tab.
        """
        return View(
               Group(
                    Group(
                         InstanceUItem('proccontroller',),
                         dock='tab',
                         label='Processing',
                         enabled_when='fd.is_loaded and  not _loading',),
                    Group(
                         InstanceUItem('fitcontroller',),
                         dock='tab',
                         label='Modelling',
                         visible_when='fitcontroller',),
                    Group(
                         InstanceUItem('plotcontroller',),
                         label='Plotting',
                         enabled_when='fd.is_loaded and not _loading',),
                    layout='tabbed',
                    show_labels=False,
                    springy=True,)
                ,)

    #===========================================================================
    # events
    #===========================================================================
    @on_trait_change('fd, fd.is_freq_items')
    def _fd_updated(self, objet, name, old, new):
        """The source fd has been changed. 
        
        We need to create the objets if they doesn't exist yet, or simply 
        update the controllers
        
        """  
        if name == 'fd':
            if self.fd.is_loaded:
                logger.debug("controller: source fd has changed")
                self.proccontroller = ProcController(fd=self.fd)
                self.fitcontroller = FitController(fd=self.fd)
                self.plotcontroller = PlotController(fd=self.fd,
                                                        viewplot=self.viewplot)
                
if __name__ == '__main__':
    pass


#! /usr/bin/env python
#===============================================================================
# processing.processbase
#===============================================================================
# Copyright (C) 2012-2013 Christian Fernandez
#       Laboratoire Catalyse et Spectrochimie, Caen, France.  
#       christian.fernandez@ensicaen.fr
# This software is governed by the CeCILL-B license under French law 
# and abiding by the rules of distribution of free software.  
# You can  use, modify and/ or redistribute the software under 
# the terms of the CeCILL-B license as circulated by CEA, CNRS and INRIA
# at the following URL "http://www.cecill.info".
# See Licence.txt in the main masai source directory
#===============================================================================

""" docstring """

import numpy as np

# low level functions
#---------------------
def _pk(data, p0=0.0, p1=0.0):
    """
    phase data
    p0 and p1 are in degrees
    """
    size = data.shape[-1]
    p0 = np.deg2rad(p0)
    p1 = np.deg2rad(p1)   
    phase = p0 + p1 * np.arange(float(size)) / float(size)
    return data * np.exp(1.0j * phase).astype(data.dtype)  

def _em(data, lb=0.0):
    """
    Exponential multiplication
    """
    return data * np.exp(-np.pi * np.arange(data.shape[-1]) * lb).astype(data.dtype)

def _zf(data, size):
    """
    zerofilling
    """
    shape = list(data.shape)
    shape[-1] = size-shape[-1]
    new = np.zeros(shape).astype(data.dtype)
    return np.hstack((data,new))

def _ft(data, si=None):
    """
    Fourier transform (first point is divided by two) and to get unity integrals the results of _ft is divided by si/2
    """
    if si is None: 
        si = data.shape[-1]
    data[0] = data[0]/2.
    return np.fft.fftshift(np.fft.fft(data, si, axis=-1), -1)/float(si/2)

def _ift(data, si=None):
    """
    Inverse fourier transform (inverse operation with respect to _ft)
    """
    if si is None: 
        si = data.shape[-1]
    ndata = np.fft.ifft(np.fft.ifftshift(data, -1), si, axis=-1)*float(si/2)
    ndata[0]=ndata[0]*2.
    return ndata

if __name__ == '__main__':
    
    pass


#TODO: replace this by my own files

#===============================================================================
# Copyright Notice and Statement for the nmrglue Project
# 
# 
# Copyright (c) 2010 Jonathan J. Helmus
# http://code.google.com/p/nmrglue/
# All rights reserved.
# 
# 
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:
# 
# 
# a. Redistributions of source code must retain the above copyright
#   notice, this list of conditions and the following disclaimer.
# 
# 
# b. Redistributions in binary form must reproduce the above copyright
#   notice, this list of conditions and the following disclaimer in the
#   documentation and/or other materials provided with the
#   distribution.
# 
# 
# c. Neither the name of the author nor the names of contributors may 
#   be used to endorse or promote products derived from this software 
#   without specific prior written permission.
# 
# 
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#===============================================================================

"""
Functions for reading and writing Bruker binary (ser/fid) files, Bruker 
JCAMP-DX parameter (acqus) files, and Bruker pulse program (pulseprogram) 
files.
"""
__developer_info__ = """
Bruker file format information
^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^

Bruker binary files (ser/fid) store data as an array of int32s whose endiness 
is determinded by the parameter BYTORDA (1 = big endian, 0 = little endian).
Typically the direct dimension is digitally filtered. The exact method of
removing this filter is unknown but an approximation is avaliable.

Bruker JCAMP-DX files (acqus, etc) are text file which are described by the 
`JCAMP-DX standard <http://www.jcamp-dx.org/>`_.  Bruker parameters are 
prefixed with a '$'.

Bruker pulseprogram files are text files described in various Bruker manuals.
Of special important are lines which describe external variable assignments 
(surrounded by "'s), loops (begin with lo), phases (contain ip of dp) or 
increments (contain id, dd, ipu or dpu).  These lines are parsed when reading
the file with nmrglue.

"""

import os
from warnings import warn

import numpy as np

from . import fileiobase
from masai.processing.processbase import _ft, _ift

# data creation

def create_data(data):
    """ 
    Create a bruker data array (recast into a complex128 or int32)
    """
    if np.iscomplexobj(data):
        return np.array(data, dtype='complex128')
    else:
        return np.array(data, dtype='int32')


# universal dictionary functions

def guess_udic(dic, data):
    """ 
    Guess parameters of universal dictionary from dic, data pair.

    Parameters
    ----------
    dic : dict
        Dictionary of Bruker parameters.
    data : ndarray
        Array of NMR data.

    Returns
    -------
    udic : dict
        Universal dictionary of spectral parameters.

    """
    # XXX if pprog, acqus are in dic use them 
    
    # create an empty universal dictionary
    udic = fileiobase.create_blank_udic(data.ndim)

    # update default values
    for i in xrange(data.ndim):
        udic[i]["size"] = data.shape[i]

    # create an empty universal dictionary
    udic = fileiobase.create_blank_udic(data.ndim)

    # update default values
    for i in xrange(data.ndim):
        udic[i]["size"] = data.shape[i]
        udic[i]['complex'] = np.iscomplexobj(data)
        udic[i]['encoding'] = 'direct'
        dim = ''
        if i > 0: dim = str(i + 1)
        acq = dic["acqu%ss" % dim]
        pro = dic["proc%ss" % dim]
        udic[i]['sw'] = acq['SW_h']
        udic[i]['obs'] = pro["SF"]
        udic[i]['car'] = acq["SFO1"]
        nuc = acq["NUCLEUS"]
        if nuc == "off":
            nuc = acq["NUC1"]
        udic[i]['label'] = nuc

    return udic

def create_dic(udic):
    """ 
    Create a Bruker parameter dictionary from a universal dictionary.

    Parameters
    ----------
    udic : dict
        Universal dictionary of spectral parameters.

    Returns
    -------
    dic : dict
        Dictionary of Bruker parameters.

    """
    ndim = udic['ndim']

    # determind the size in bytes
    if udic[ndim - 1]["complex"]:
        bytes = 8
    else:
        bytes = 4

    for k in xrange(ndim): 
        bytes *= udic[k]["size"]

    dic = {"FILE_SIZE":bytes}

    # create the pprog dictionary parameter
    dic["pprog"] = {'incr': [[], [1]] * (ndim*2-2), 
                    'loop': [2] * (ndim * 2 - 2),
                    'ph_extra': [[]] * (ndim * 2 - 2),
                    'phase': [[]] * (ndim * 2 - 2),
                    'var': {}}
    
    # create acqus and procs dictionary parameters and fill in loop sizes
    dic['acqus'] = create_acqus_dic(udic[ndim - 1], direct=True)
    dic['procs'] = create_procs_dic(udic[ndim - 1])
    if ndim >= 2:
        dic["acqu2s"] = create_acqus_dic(udic[ndim - 2])
        dic["proc2s"] = create_procs_dic(udic[ndim - 2])
        dic["pprog"]["loop"][1] = udic[ndim - 2]["size"] / 2
    if ndim >= 3:
        dic["acqu3s"] = create_acqus_dic(udic[ndim - 3])
        dic["proc3s"] = create_procs_dic(udic[ndim - 3])
        dic["pprog"]["loop"][3] = udic[ndim - 3]["size"] / 2
    if ndim >= 4:
        dic["acqu4s"] = create_acqus_dic(udic[ndim - 4])
        dic["proc4s"] = create_procs_dic(udic[ndim - 4])
        dic["pprog"]["loop"][5] = udic[ndim - 4]["size"] / 2
    
    return dic

def create_acqus_dic(adic, direct=False):
    """
    Create a Bruker acqus dictionary from an Universal axis dictionary.  
    Set direct=True for direct dimension.
    """
    if adic["complex"]:
        AQ_mod = 3
        if direct:
            TD = int(np.ceil(adic["size"] / 256.) * 256 ) * 2
        else:
            TD = adic["size"]
    else:
        AQ_mod = 1
        if direct:
            TD = int(np.ceil(adic["size"] / 256.) * 256)
        else:
            TD = adic["size"]

    s = '##NMRGLUE automatically created parameter file'
    return {'_comments':[], '_coreheader':[s], 'AQ_mod':AQ_mod, 'TD':TD}

def create_procs_dic(adic, direct=False):
    """
    Create a Bruker procs dictionary from an Universal axis dictionary.  
    """
    s = '##NMRGLUE automatically created parameter file'
    return {'_comments':[], '_coreheader':[s], 'SF':0 }


# Global read/write function and related utilities

def read(dir=".", bin_file=None, acqus_files=None, procs_files=None, pprog_file=None,
        shape=None, cplex=None, big=None, read_prog=True, read_acqus=True, read_procs=True):
    """ 
    Read Bruker files from a directory.

    Parameters
    ----------
    dir : str
        Directory to read from.
    bin_file : str, optional
        Filename of binary file in directory. None uses standard files.
    acqus_files : list, optional
        List of filename(s) of acqus parameter files in directory. None uses
        standard files.
    procs_files : list, optional
        List of filename(s) of procs parameter files in pdata/1/ directory. None uses
        standard files.
    pprog_file : str, optional
        Filename of pulse program in directory. None uses standard files.
    shape : tuple, optional
        Shape of resulting data.  None will guess the shape from the spectral
        parameters.
    cplex : bool, optional
        True is direct dimension is complex, False otherwise. None will guess
        quadrature from spectral parameters.
    big : bool or None, optional
        Endiness of binary file. True of big-endian, False for little-endian,
        None to determine endiness from acqus file(s).
    read_pprog : bool, optional
        True to read pulse program, False prevents reading.
    read_acqus : bool, optional
        True to read acqus files(s), False prevents reading.
    read_procs : bool, optional
        True to read procs files(s), False prevents reading.

    Returns
    -------
    dic : dict
        Dictionary of Bruker parameters.
    data : ndarray
        Array of NMR data.

    See Also
    --------
    read_lowmem : Low memory reading of Bruker files.
    write : Write Bruker files.

    """
    if os.path.isdir(dir) != True:
        raise IOError, "directory %s does not exist" % (dir)

    # determind parameter automatically
    if bin_file == None:
        if os.path.isfile(os.path.join(dir, "fid")):
            bin_file = "fid"
        elif os.path.isfile(os.path.join(dir, "ser")):
            bin_file = "ser"
        else:
            raise IOError,"No Bruker binary file could be found in %s" % (dir)

    if acqus_files == None:
        acqus_files = []
        for f in ["acqus", "acqu2s", "acqu3s", "acqu4s"]:
            if os.path.isfile(os.path.join(dir, f)):
                acqus_files.append(f)

    if procs_files == None:
        procs_files = []
        for f in ["procs", "proc2s", "proc3s", "proc4s"]:
            if os.path.isfile(os.path.join(dir, "pdata/1", f)):
                procs_files.append(f)
                
    if pprog_file == None:
        pprog_file = "pulseprogram"

    # create an empty dictionary
    dic = dict()

    # read the acqus_files and add to the dictionary
    if read_acqus:
        for f in acqus_files:
            dic[f] = read_jcamp(os.path.join(dir, f))

    # read the procs_files and add to the dictionary
    if read_procs:
        for f in procs_files:
            dic[f] = read_jcamp(os.path.join(dir, "pdata/1", f))
            
    # read the pulse program and add to the dictionary
    if read_prog:
        dic["pprog"] = read_pprog(os.path.join(dir, pprog_file))

    # determind file size and add to the dictionary
    dic["FILE_SIZE"] = os.stat(os.path.join(dir, bin_file)).st_size

    # determind shape and complexity for direct dim if needed
    if shape == None or cplex == None:
        gshape, gcplex = guess_shape(dic)
        if gcplex == True:    # divide last dim by 2 if complex
            t = list(gshape)
            t[-1] = t[-1] / 2
            gshape = tuple(t)
    if shape == None:
        shape = gshape
    if cplex == None:
        cplex = gcplex
   
    # determind endianness (assume little-endian unless BYTORDA is 1)
    if big == None:
        big = False # default value
        if "acqus" in dic and "BYTORDA" in dic["acqus"]:
            if dic["acqus"]["BYTORDA"] == 1:
                big = True
            else:
                big = False

    # read the binary file
    f = os.path.join(dir, bin_file)
    null, data = read_binary(f, shape=shape, cplex=cplex, big=big)
    return dic, data


def read_lowmem(dir=".", bin_file=None, acqus_files=None, procs_files=None, pprog_file=None,
        shape=None, cplex=None, big=None, read_prog=True, read_acqus=True, read_procs=True):
    """ 
    Read Bruker files from a directory using minimal amounts of memory.

    See :py:func:`read` for Parameters.

    Returns
    -------
    dic : dict
        Dictionary of Bruker parameters.
    data : array_like
        Low memory object which can access NMR data on demand.

    See Also
    --------
    read : Read Bruker files. 
    write_lowmem : Write Bruker files using minimal amounts of memory.

    """

    if os.path.isdir(dir) != True:
        raise IOError, "directory %s does not exist" % (dir)

    # determind parameter automatically
    if bin_file == None:
        if os.path.isfile(os.path.join(dir, "fid")):
            bin_file = "fid"
        elif os.path.isfile(os.path.join(dir, "ser")):
            bin_file = "ser"
        else:
            raise IOError, "no Bruker binary file could be found in %s" % (dir)

    if acqus_files == None:
        acqus_files = []
        for f in ["acqus", "acqu2s", "acqu3s", "acqu4s"]:
            if os.path.isfile(os.path.join(dir, f)):
                acqus_files.append(f)

    if procs_files == None:
        procs_files = []
        for f in ["procs", "proc2s", "proc3s", "proc4s"]:
            if os.path.isfile(os.path.join(dir, "pdata/1",f)):
                acqus_files.append(f)

    if pprog_file == None:
        pprog_file = "pulseprogram"

    # create an empty dictionary
    dic = dict()

    # read the acqus_files and add to the dictionary
    if read_acqus:
        for f in acqus_files:
            dic[f] = read_jcamp(os.path.join(dir, f))

    # read the procs_files and add to the dictionary
    if read_procs:
        for f in procs_files:
            dic[f] = read_jcamp(os.path.join(dir, "pdata/1",f))

    # read the pulse program and add to the dictionary
    if read_prog:
        dic["pprog"] = read_pprog(os.path.join(dir, pprog_file))

    # determind file size and add to the dictionary
    dic["FILE_SIZE"] = os.stat(os.path.join(dir, bin_file)).st_size

    # determind shape and complexity for direct dim if needed
    if shape == None or cplex == None:
        gshape, gcplex = guess_shape(dic)
        if gcplex == True:    # divide last dim by 2 if complex
            t = list(gshape)
            t[-1] = t[-1] / 2
            gshape = tuple(t)
    if shape == None:
        shape = gshape
    if cplex == None:
        cplex = gcplex
   
    # determind endianness (assume little-endian unless BYTORDA is 1)
    if big == None:
        big = False # default value
        if "acqus" in dic and "BYTORDA" in dic["acqus"]:
            if dic["acqus"]["BYTORDA"] == 1:
                big = True
            else:
                big = False

    # read the binary file
    f = os.path.join(dir, bin_file)
    null, data = read_binary_lowmem(f, shape=shape, cplex=cplex, big=big)
    return dic, data

#TODO: modify to also write procs file
def write(dir, dic, data, bin_file=None, acqus_files=None, pprog_file=None,
        overwrite=False, big=None, write_prog=True, write_acqus=True):
    """ 
    Write Bruker files to disk.

    Parameters
    ----------
    dir : str
        Directory to write files to.
    dir : dict
        Dictionary of Bruker parameters.
    data : array_like
        Array of NMR data
    bin_file : str, optional
        Filename of binary file in directory. None uses standard files.
    acqus_files : list, optional
        List of filename(s) of acqus parameter files in directory. None uses
        standard files.
    pprog_file : str, optional
        Filename of pulse program in directory. None uses standard files.
    overwrite : bool, optional
        Set True to overwrite files, False will raise a Warning if files 
        exist.
    big : bool or None, optional
        Endiness of binary file. True of big-endian, False for little-endian,
        None to determine endiness from Bruker dictionary.
    write_pprog : bool, optional
        True to write the pulse program file, False prevents writing.
    write_acqus : bool, optional
        True to write the acqus files(s), False prevents writing.
    
    See Also
    --------
    write_lowmem : Write Bruker files using minimal amounts of memory.
    read : Read Bruker files.

    """
    # determind parameters automatically
    if bin_file == None:
        if data.ndim == 1:
            bin_file = "fid"
        else:
            bin_file = "ser"

    if acqus_files == None:
        acq = ["acqus", "acqu2s", "acqu3s", "acqu4s"]
        acqus_files = [k for k in acq if dic.has_key(k)]

    if pprog_file == None:
        pprog_file = "pulseprogram"

    # write out the acqus files
    if write_acqus:
        for f in acqus_files:
            write_jcamp(dic[f], os.path.join(dir, f), overwrite=overwrite)

    # write out the pulse program
    if write_prog:
        write_pprog(os.path.join(dir, pprog_file), dic["pprog"], 
                        overwrite=overwrite)
    
    # determind endianness (assume little-endian unless BYTORDA is 1)
    if big == None:
        big = False # default value
        if "acqus" in dic and "BYTORDA" in dic["acqus"]:
            if dic["acqus"]["BYTORDA"] == 1:
                big = True
            else:
                big = False

    # write out the binary data
    bin_full = os.path.join(dir, bin_file)
    write_binary(bin_full, dic, data, big=big, overwrite=overwrite)
    return

#TODO: modify to also write procs file
def write_lowmem(dir, dic, data, bin_file=None, acqus_files=None, 
        pprog_file=None, overwrite=False, big=None, write_prog=True, 
        write_acqus=True):
    """ 
    Write Bruker files using minimal amounts of memory (trace by trace).

    See :py:func:`write` for Parameters.

    See Also
    --------
    write : Write Bruker files.
    read_lowmem : Read Bruker files using minimal amounts of memory.

    """
    # determind parameters automatically
    if bin_file == None:
        if data.ndim == 1:
            bin_file = "fid"
        else:
            bin_file = "ser"

    if acqus_files == None:
        acq = ["acqus", "acqu2s", "acqu3s", "acqu4s"]
        acqus_files = [k for k in acq if dic.has_key(k)]

    if pprog_file == None:
        pprog_file = "pulseprogram"

    # write out the acqus files
    if write_acqus:
        for f in acqus_files:
            write_jcamp(dic[f], os.path.join(dir, f), overwrite=overwrite)

    # write out the pulse program
    if write_prog:
        write_pprog(os.path.join(dir, pprog_file), dic["pprog"],
                    overwrite=overwrite)

    # determind endianness (assume little-endian unless BYTORDA is 1)
    if big == None:
        big = False # default value
        if "acqus" in dic and "BYTORDA" in dic["acqus"]:
            if dic["acqus"]["BYTORDA"] == 1:
                big = True
            else:
                big = False

    # write out the binary data
    bin_full = os.path.join(dir, bin_file)
    write_binary_lowmem(bin_full, dic, data, big=big, overwrite=overwrite)
    return

def guess_shape(dic):
    """
    Determine data shape and complexity from Bruker dictionary.

    Returns
    -------
    shape : tuple
        Shape of data in Bruker binary file (R+I for all dimensions).
    cplex : bool 
        True for complex data in last (direct) dimension, False otherwise.

    ...Note:: Modified from the nmrglue routine. Read only well formed file
    and do not try to guess too much.
    
    """
    # determine complexity of last (direct) dimension
    try:
        aq_mod = dic["acqus"]["AQ_mod"]
    except KeyError:
        return (1,), False
    
    if aq_mod == 0 or aq_mod == 2:
        cplex = False
    elif aq_mod == 1 or aq_mod == 3:
        cplex = True
    else:
        warn("Unknown Aquisition Mode")
        return (1,), False
    
    # file size
    try:
        fsize = dic["FILE_SIZE"]
    except KeyError:
        warn("cannot determine shape do to missing FILE_SIZE key")
        return (1,), False
    
    # shape of the data
    parmode = -1
    try:
        parmode = dic["acqus"]["PARMODE"]
    except KeyError:
        warn("cannot determine the PARMODE parameter: you may correct this in the acqus file.")
        return (1,), False
    
    
    shape = range(parmode+1)
    for i in range(1, parmode+1):
        if dic.has_key("acqu%ss"%(i+1)):
            if dic["acqu%ss"%(i+1)].has_key("TD"):
                shape[i]=float(dic["acqu%ss"%(i+1)]["TD"])
            else:
                warn("Parameter TD is missing in acqu%ss: you may correct this in the file.")
                return (1,), False
            
    # last (direct) dimension is given by "TD" parameter in acqus file
    # rounded up to nearest 256
    shape[0]=int(np.ceil( float(dic["acqus"]["TD"]) / 256.)*256.)
    shape.reverse()
    
    return tuple([int(i) for i in shape if i >= 2]), cplex


#    # last (direct) dimension is given by "TD" parameter in acqus file
#    # rounded up to nearest 256
#    # next-to-last dimension may be given by "TD" in acqu2s. In 3D+ data
#    # this is often the sum of the indirect dimensions
#    shape  = [0, 0, td2, int(np.ceil(td0 / 256.)*256.)]
#    
#    # additional dimension given by data size
#    if shape[2] != 0 and shape[3] != 0:
#        shape[1] = fsize / (shape[3] * shape[2] * 4)
#        shape[0] = fsize / (shape[3] * shape[2] * 16 * 4)
#    
#    # if there in no pulse program parameters in dictionary return currect
#    # shape after removing zeros
#    if "pprog" not in dic or "loop" not in dic["pprog"]:
#        return tuple([int(i) for i in shape if i >= 1]), cplex
#
#    # if pulseprogram dictionary is missing loop or incr return current shape
#    pprog = dic["pprog"]
#    if "loop" not in pprog or "incr" not in pprog:
#        return tuple([int(i) for i in shape if i >= 1]), cplex
#
#    # determine indirect dimension sizes from pulseprogram parameters
#    loop = pprog["loop"]
#    loopn = len(loop)       # number of loops
#    li = [len(i) for i in pprog["incr"]] # length of incr lists
#
#    # replace td0,td1,td2,td3 in loop list
#    rep = {'td0' : td0, 'td1' : td1, 'td2' : td2, 'td3' : td3}
#    for i, v in enumerate(loop):
#        if v in rep.keys():
#            loop[i] = rep[v]
#             
#    # size of indirect dimensions based on number of loops in pulse program
#    # there are two kinds of loops, active and passive.
#    # active loops are from indirect dimension increments, the corresponding
#    # incr lists should have non-zero length and the size of the dimension
#    # is twice that of the active loop size.
#    # passive loops are from phase cycles and similar elements, these should 
#    # have zero length incr lists and should be of length 2.  
#
#    # The following checks for these and updates the indirect dimension
#    # if the above is found.
#    if loopn == 1:    # 2D with no leading passive loops
#        if li[0] != 0:
#            shape[2] = loop[0]
#            shape = shape[-2:]
#    
#    elif loopn == 2:  # 2D with one leading passive loop
#        if loop[0] == 2 and li[0] == 0 and li[1] != 0:
#            shape[2] = 2 * loop[1]
#            shape = shape[-2:]
#
#    elif loopn == 3:  # 2D with two leading passive loops
#        if (loop[0] == 2 and loop[1] == 2 and li[0] == 0 and li[1] == 0 
#            and li[2] != 0):
#            shape[2] = 2 * loop[2]
#            shape = shape[-2:]
#
#    elif loopn == 4:  # 3D with one leading passive loop for each indirect dim
#        if loop[0] == 2 and li[0] == 0 and li[1] != 0:
#            shape[2] = 2 * loop[1]
#        if loop[2] == 2 and li[2] == 0 and li[3] != 0:
#            shape[1] = 2 * loop[3]
#            shape = shape[-3:]
#
#    elif loopn == 5:  # 3D with two/one leading passive loops
#        if loop[1] == 2 and li[0] == 0 and li[1] == 0 and li[2] != 0:
#            shape[2] = 2 * loop[2]
#        if loop[3] == 2 and li[0] == 0 and li[3] == 0 and li[4] != 0:
#            shape[1] = 2 * loop[4]
#            shape = shape[-3:]
#
#    elif loopn == 6:  # 4D with one leading passive loop for each indirect dim
#        if loop[0] == 2 and li[0] == 0 and li[1] != 0:
#            shape[2] = 2 * loop[1]
#        if loop[2] == 2 and li[2] == 0 and li[3] != 0:
#            shape[1] = 2 * loop[3]
#        if loop[4] == 2 and li[4] == 0 and li[5] != 0:
#            shape[0] = 2 * loop[5]
#            
#    elif loopn == 7:
#        if loop[1] == 2 and li[0] == 0 and li[1] == 0 and li[2] != 0:
#            shape[2] = 2 * loop[2]
#        if loop[3] == 2 and li[0] == 0 and li[3] == 0 and li[4] != 0:
#            shape[1] = 2 * loop[4]
#        if loop[5] == 2 and li[0] == 0 and li[5] == 0 and li[6] != 0:
#            shape[0] = 2 * loop[6]
#
#    return tuple([int(i) for i in shape if i >= 2]), cplex

# Bruker binary (fid/ser) reading and writing


def read_binary(filename, shape=(1), cplex=True, big=True):
    """ 
    Read Bruker binary data from file and return dic,data pair

    If data cannot be reshaped as described a 1D representation of the data 
    will be returned after printing a warning message.
    
    Parameters
    ----------
    filename : str 
        Filename of Bruker binary file.
    shape : tuple
        Tuple describing shape of resulting data.
    cplex : bool
        Flag indicating if direct dimension is complex.
    big : bool
        Endianness of binary file, True for big-endian, False for 
        little-endian.

    Returns
    -------
    dic : dict
        Dictionary containing "FILE_SIZE" key and value.
    data : ndarray
        Array of raw NMR data.

    See Also
    --------
    read_binary_lowmem : Read Bruker binary file using minimal memory.

    """
    # open the file and get the data
    f = open(filename, 'rb')
    data = get_data(f, big=big)

    # complexify if needed
    if cplex:
        data = complexify_data(data)

    # create dictionary
    dic = {"FILE_SIZE" : os.stat(filename).st_size}

    # reshape if possible
    try:
        return dic, data.reshape(shape)

    except ValueError:
        warn(str(data.shape) + "cannot be shaped into" + str(shape))
        return dic, data

def read_binary_lowmem(filename, shape=(1), cplex=True, big=True):
    """ 
    Read Bruker binary data from file using minimal memory.

    Raises ValueError if shape does not agree with file size.
    See :py:func:`read_binary` for Parameters.
    
    Returns
    -------
    dic : dict
        Dictionary containing "FILE_SIZE" key and value.
    data : array_like
        Low memory object which can access NMR data on demand.

    See Also
    --------
    read_binary: Read Bruker binary file.

    """
    # create dictionary
    dic = {"FILE_SIZE" : os.stat(filename).st_size}
    data = bruker_nd(filename, shape, cplex, big)
    return dic, data

def write_binary(filename, dic, data, overwrite=False, big=True):
    """ 
    Write Bruker binary data to file.

    Parameters
    ----------
    filename : str
        Filename to write to.
    dic : dict
        Dictionary of Bruker parameters.
    data : ndarray
        Array of NMR data.
    overwrite : bool
        True to overwrite files, False will raise a Warning if file exists.
    big : bool
        Endiness to write binary data with True of big-endian, False for
        little-endian.
    
    See Also
    --------
    write_binary_lowmem : Write Bruker binary data using minimal memory.

    """
    # open the file for writing
    f = fileiobase.open_towrite(filename, overwrite=overwrite)
    
    # convert objec to an array if it is not already one...
    if type(data) != np.ndarray:
        data = np.array(data)

    if np.iscomplexobj(data):
        put_data(f, uncomplexify_data(data), big)
    else:
        put_data(f, data, big)
    f.close()
    return

def write_binary_lowmem(filename, dic, data, overwrite=False, big=True):
    """ 
    Write Bruker binary data to file using minimal memory (trace by trace).

    See :py:func:`write_binary` for Parameters.

    See Also
    --------
    write_binary : Write Bruker binary data to file.
    
    """
    # open the file for writing
    f = fileiobase.open_towrite(filename, overwrite=overwrite)

    cplex = np.iscomplexobj(data)

    # write out file trace by trace
    for tup in np.ndindex(data.shape[:-1]):
        trace = data[tup]
        if cplex:
            put_data(f, uncomplexify_data(trace), big)
        else:
            put_data(f, trace, big)
    f.close()
    return


# lowmemory ND object

class bruker_nd(fileiobase.data_nd):
    """
    Emulate a ndarray objects without loading data into memory for low memory 
    reading of Bruker fid/ser files.

    * slicing operations return ndarray objects.
    * can iterate over with expected results.
    * transpose and swapaxes methods create a new objects with correct axes
      ordering.
    * has ndim, shape, and dtype attributes.

    Parameters
    ----------

    filename : str
        Filename of Bruker binary file.
    fshape : tuple
        Shape of NMR data.
    cplex : bool
        Flag indicating if direct dimension is complex.
    big : bool
        Endianess of data.  True for big-endian, False for little-endian.
    order : tuple
        Ordering of axis against file.

    """

    def __init__(self, filename, fshape, cplex, big, order=None):
        """
        Create and set up object.
        """
        
        # check that size is correct
        pts = reduce(lambda x, y: x*y, fshape)
        if cplex:
            if os.stat(filename).st_size != pts * 4 * 2:
                raise ValueError("shape does not agree with file size")
        else:
            if os.stat(filename).st_size != pts * 4:
                raise ValueError("shape does not agree with file size")

        # check order
        if order == None:
            order = range(len(fshape))
        
        # finalize
        self.filename = filename
        self.fshape = fshape
        self.cplex = cplex
        self.big = big
        self.order = order

        if self.cplex:
            self.dtype = np.dtype("complex128")
        else:
            self.dtype = np.dtype("int32")
        
        self.__setdimandshape__()   # set ndim and shape attributes

    def __fcopy__(self, order):
        """ 
        Create a copy
        """
        n = bruker_nd(self.filename, self.fshape, self.cplex, self.big, order)
        return n

    def __fgetitem__(self, slices):
        """
        return ndarray of selected values

        slices is a well formatted tuple of slices
        """
        # seperate the last slice from the first slices
        lslice = slices[-1]
        fslice = slices[:-1]

        # and the same for fshape
        lfshape = self.fshape[-1]
        ffshape = self.fshape[:-1]

        # find the output size and make a in/out nd interator
        osize, nd_iter = fileiobase.size_and_ndtofrom_iter(ffshape, fslice)
        osize.append(len(range(lfshape)[lslice]))

        # create an empty array to store the selected slices
        out = np.empty(tuple(osize), dtype=self.dtype)

        f = open(self.filename, 'rb')

        # read in the data trace by trace
        for out_index, in_index in nd_iter:

            # determine the trace number from the index
            ntrace = fileiobase.index2trace_flat(ffshape, in_index)

            # seek to the correct place in the file
            if self.cplex:
                ts = ntrace * lfshape * 2 * 4
                f.seek(ts)
                trace = get_trace(f, lfshape * 2, self.big)
                trace = complexify_data(trace)
            else:
                ts = ntrace * lfshape * 2
                f.seek(ts)
                trace = get_trace(f, lfshape, self.big)
 
            # save to output
            out[out_index] = trace[lslice]
        
        return out

# binary get/put functions

def get_data(f, big):
    """ 
    Get binary data from file object with given endiness
    """
    if big == True:
        return np.frombuffer(f.read(), dtype='>i4')
    else:
        return np.frombuffer(f.read(), dtype='<i4')

def put_data(f, data, big=True):
    """ 
    Put data to file object with given endiness
    """
    if big:
        f.write(data.astype('>i4').tostring())
    else:
        f.write(data.astype('<i4').tostring())
    return

def get_trace(f, num_points, big):
    """ 
    Get trace of num_points from file with given endiness
    """
    if big == True:
        bsize = num_points * np.dtype('>i4').itemsize
        return np.frombuffer(f.read(bsize), dtype='>i4')
    else:
        bsize = num_points * np.dtype('<i4').itemsize
        return np.frombuffer(f.read(bsize), dtype='<i4')

# data manipulation functions

def complexify_data(data):
    """ 
    Complexify data packed real, imag.
    """
    return data[..., ::2] + data[..., 1::2]*1.j

def uncomplexify_data(data_in):
    """ 
    Uncomplexify data (pack real,imag) into a int32 array
    """
    size = list(data_in.shape)
    size[-1] = size[-1] * 2
    data_out = np.empty(size, dtype="int32")
    data_out[..., ::2]  = data_in.real
    data_out[..., 1::2] = data_in.imag
    return data_out


# digital filter functions

# Table of points to frequency shift Bruker data to remove digital filter
# (Phase is 360 degrees * num_pts)
# This table is an 'un-rounded' version base on the table by
# W.M. Westler and F. Abildgaard's offline processing note, online at:
# http://www.boc.chem.uu.nl/static/local/prospectnd/dmx_digital_filters.html
# and the updated table with additional entries at:
# http://sbtools.uchc.edu/help/nmr/nmr_toolkit/bruker_dsp_table.asp

# The rounding in the above tables appear to be based on k / (2*DECIM)
# for example 2 : 44.75   = 44 + 3/4
#             4 : 66.625  = 66 + 5/8
#             8 : 68.563 ~= 68 + 9/16 = 68.5625
# Using this the un-rounded table was created by checking possible unrounded
# fracions which would round to those in the original table.

bruker_dsp_table = {
    10: { 
        2    : 44.75,
        3    : 33.5,
        4    : 66.625,
        6    : 59.083333333333333,
        8    : 68.5625,
        12   : 60.375,
        16   : 69.53125,
        24   : 61.020833333333333,
        32   : 70.015625,
        48   : 61.34375,
        64   : 70.2578125,
        96   : 61.505208333333333,
        128  : 70.37890625,
        192  : 61.5859375,
        256  : 70.439453125,
        384  : 61.626302083333333,
        512  : 70.4697265625,
        768  : 61.646484375,
        1024 : 70.48486328125,
        1536 : 61.656575520833333,
        2048 : 70.492431640625,
        },
    11: {
        2    : 46.,
        3    : 36.5,
        4    : 48.,
        6    : 50.166666666666667,
        8    : 53.25,
        12   : 69.5,
        16   : 72.25,
        24   : 70.166666666666667,
        32   : 72.75,
        48   : 70.5,
        64   : 73.,
        96   : 70.666666666666667,
        128  : 72.5,
        192  : 71.333333333333333,
        256  : 72.25,
        384  : 71.666666666666667,
        512  : 72.125,
        768  : 71.833333333333333,
        1024 : 72.0625,
        1536 : 71.916666666666667,
        2048 : 72.03125
        },
    12: {
        2    : 46. ,
        3    : 36.5,
        4    : 48.,
        6    : 50.166666666666667,
        8    : 53.25,
        12   : 69.5,
        16   : 71.625,
        24   : 70.166666666666667,
        32   : 72.125,
        48   : 70.5,
        64   : 72.375,
        96   : 70.666666666666667,
        128  : 72.5,
        192  : 71.333333333333333,
        256  : 72.25,
        384  : 71.666666666666667,
        512  : 72.125,
        768  : 71.833333333333333,
        1024 : 72.0625,
        1536 : 71.916666666666667,
        2048 : 72.03125
        },
    13: {
        2    : 2.75, 
        3    : 2.8333333333333333,
        4    : 2.875,
        6    : 2.9166666666666667,
        8    : 2.9375,
        12   : 2.9583333333333333,
        16   : 2.96875,
        24   : 2.9791666666666667,
        32   : 2.984375,
        48   : 2.9895833333333333,
        64   : 2.9921875,
        96   : 2.9947916666666667
        } 
    }

def remove_digital_filter(dic, data):
    """
    Remove the digial filter from Bruker data.

    Parameters
    ----------
    dic : dict
        Dictionary of Bruker parameters.
    data : ndarray
        Array of NMR data to remove digital filter from.

    Returns
    -------
    ndata : ndarray
        Array of NMR data with digital filter removed

    See Also
    ---------
    rm_dig_filter : Remove digital filter by specifying parameters.

    """
    if 'acqus' not in dic:
        raise ValueError("dictionary does not contain acqus parameters") 
    
    if 'DECIM' not in dic['acqus']:
        raise ValueError("dictionary does not contain DECIM parameter")
    decim = dic['acqus']['DECIM']
    
    if 'DSPFVS' not in dic['acqus']:
        raise ValueError("dictionary does not contain DSPFVS parameter")
    dspfvs = dic['acqus']['DSPFVS']

    if 'GRPDLY' not in dic['acqus']:
        grpdly = 0
    else:
        grpdly = dic['acqus']['GRPDLY']
 
    return rm_dig_filter(data, decim, dspfvs, grpdly)
    
def rm_dig_filter(data, decim, dspfvs, grpdly=0):
    """
    Remove the digital filter from Bruker data.

    Parameters
    ----------
    data : ndarray
        Array of NMR data to remove digital filter from.
    decim : int 
        Decimation rate (Bruker DECIM parameter).
    dspfvs : int    
        Firmware version (Bruker DSPFVS parameter).
    grpdly : float, optional
        Group delay. (Bruker GRPDLY parameter). When non-zero decom and 
        dspfvs are ignored.

    Returns
    -------
    ndata : ndarray
        Array of NMR data with digital filter removed.

    See Also
    --------
    remove_digital_filter : Remove digital filter using Bruker dictionary.

    """
    # This algorithm gives results similar but not exactly the same
    # as NMRPipe.  It was worked out by examining sample FID converted using
    # NMRPipe against spectra shifted with nmrglue's processing functions.  
    # When a frequency shifting with a fft first (fft->first order phase->ifft)
    # the middle of the fid nearly matches NMRPipe's and the difference at the
    # beginning is simply the end of the spectra reversed.  A few points at 
    # the end of the spectra are skipped entirely. 
    # -jjh 2010.12.01

    # The algorithm is as follows:
    # 1. FFT the data
    # 2. Apply a negative first order phase to the data.  The phase is 
    #    determined by the GRPDLY parameter or found in the DSPFVS/DECIM 
    #    loopup table.
    # 3. Inverse FFT  
    # (these first three steps are a frequency shift with a FFT first, fsh2)
    # 4. Round the applied first order phase up by two integers. For example
    #    71.4 -> 73, 67.8 -> 69, and 48 -> 50, this is the number of points
    #    removed from the end of the fid.
    # 5. If the size of the removed portion is greater than 6, remove the first
    #    6 points, reverse the remaining points, and add then to the beginning
    #    of the spectra.  If less that 6 points were removed, leave the FID 
    #    alone.
   
    if grpdly > 0:  # use group delay value if provided (not 0 or -1)
        phase = grpdly
    
    # determind the phase correction
    else:
        if dspfvs >= 14:    # DSPFVS greater than 14 give no phase correction.
            phase = 0.
        else:   # loop up the phase in the table
            if dspfvs not in bruker_dsp_table:
                raise ValueError("dspfvs not in lookup table")
            if decim not in bruker_dsp_table[dspfvs]:
                raise ValueError("decim not in lookup table")
            phase = bruker_dsp_table[dspfvs][decim]

    # and the number of points to remove (skip) and add to the beginning
    skip = int(np.floor(phase+2.))  # round up two integers
    add = int(max(skip-6, 0))        # 6 less, or 0

    # DEBUG 
    print "4 phase: %f, skip: %i add: %i"%(phase,skip,add)

    # frequency shift
    #pdata = proc_base.fsh2(data, phase)
    

    # The above process give bad spectra in my case with artifact in the midle
    # so I modified it this way (this is likely due to the fsh2 procedure)
    pdata = _ft(data)
    s = float(pdata.shape[-1])
    ph = 2.*np.pi * phase * np.arange(s) / s
    pdata = pdata * np.exp(1j * ph)
    pdata = _ift(pdata)
    
    # add points at the end of the specta to beginning
    pdata[..., :add] = pdata[..., :add]+pdata[..., :-(add+1):-1]
    
    # remove points at end of spectra
    return pdata[..., :-skip]


# JCAMP-DX functions

def read_jcamp(filename):
    """ 
    Read a Bruker JCAMP-DX file into a dictionary.

    Creates two special dictionary keys _coreheader and _comments Bruker 
    parameter "$FOO" are extracted into strings, floats or lists and assigned 
    to dic["FOO"]

    Parameters
    ----------
    filename : str
        Filename of Bruker JCAMP-DX file.

    Returns
    -------
    dic : dict
        Dictionary of parameters in file.

    See Also
    --------
    write_jcamp : Write a Bruker JCAMP-DX file.

    Notes
    -----
    This is not a fully functional JCAMP-DX reader, it is only intended
    to read Bruker acqus (and similar) files.

    """
    dic = {"_coreheader":[], "_comments":[]}    # create empty dictionary
    f = open(filename, 'rb')

    # loop until EOF
    while len(f.read(1)):
        
        f.seek(-1, os.SEEK_CUR)  # rewind 1 byte
        line = f.readline().rstrip()    # read a line

        if line[:6] == "##END=":
            #print "End of file"
            break
        elif line[:2] == "$$":
            dic["_comments"].append(line)
        elif line[:2] == "##" and line[2] != "$":
            dic["_coreheader"].append(line)
        elif line[:3] == "##$":
            key, value = parse_jcamp_line(line, f)
            dic[key] = value
        else:
            warn("Extraneous line:" + line)

    return dic

def parse_jcamp_line(line, f):
    """ 
    Parse a single JCAMP-DX line
    
    Extract the Bruker parameter name and value from a line from a JCAMP-DX
    file.  This may entail reading additional lines from the fileobj f if the 
    parameter value extends over multiple lines.
    
    """

    # extract key= text from line
    key = line[3:line.index("=")]
    text = line[line.index("=") + 1:].lstrip()

    if "<" in text:   # string
        while ">" not in text:      # grab additional text until ">" in string
            text = text + "\n" + f.readline().rstrip() 
        value = text.replace("<", "").replace(">", "")

    elif "(" in text: # array
        num = int(line[line.index("..")+2:line.index(")")]) + 1
        value = []
        rline = line[line.index(")") + 1:]

        # extract value from remainer of line
        for t in rline.split():
            if "." in t or "e" in t:
                value.append(float(t))
            else:
                value.append(int(t))
        
        # parse additional lines as necessary
        while len(value) < num:
            nline = f.readline().rstrip()
            for t in nline.split():
                if "." in t or "e" in t:
                    value.append(float(t))
                else:
                    value.append(int(t))

    elif text == "yes":
        value = True

    elif text == "no":
        value = False

    else:   # simple value
        if "." in text or "e" in text: 
            value = float(text)
        else:
            value = int(text)

    return key, value


def write_jcamp(dic, filename, overwrite=False):
    """ 
    Write a Bruker JCAMP-DX file from a dictionary

    Written file will differ slightly from Bruker's JCAMP-DX files in that all
    multi-value parameters will be written on multiple lines. Bruker is 
    inconsistent on what is written to a single line and what is not.  
    In addition line breaks may be slightly different but will always be 
    within JCAMP-DX specification.  Finally long floating point values
    may loose precision when writing.

    For example:

        ##$QS= (0..7)83 83 83 83 83 83 83 22
        
        will be written as

        ##$QS= (0..7)
        83 83 83 83 83 83 83 22

    Parameters
    ----------
    dic : dict
        Dictionary of parameters to write
    filename : str
        Filename of JCAMP-DX file to write
    overwrite : bool, optional
        True to overwrite an existing file, False will raise a Warning if the
        file already exists.

    See Also
    --------
    read_jcamp : Read a Bruker JCAMP-DX file.

    """
 
    # open the file for writing
    f = fileiobase.open_towrite(filename, overwrite=overwrite)

    # create a copy of the dictionary
    d = dict(dic)

    # remove the comments and core header key from dictionary
    comments = d.pop("_comments")
    corehdr = d.pop("_coreheader")

    # write out the core headers
    for line in corehdr:
        f.write(line)
        f.write("\n")

    # write out the comments
    for line in comments:
        f.write(line)
        f.write("\n")

    keys = d.keys()
    keys.sort()

    # write out each key,value pair
    for key in keys:
        write_jcamp_pair(f, key, d[key])

    # write ##END= and close the file
    f.write("##END=")
    f.close()


def write_jcamp_pair(f, key, value):
    """ 
    Write out a line of a JCAMP file.

    a line might actually be more than one line of text for arrays.
    """

    # the parameter name and such
    line = "##$" + key + "= "

    if type(value) == float or type(value) == int:  # simple numbers
        line = line + repr(value)

    elif type(value) == str:        # string
        line = line + "<" + value + ">"

    elif type(value) == bool:   # yes or no
        if value:
            line = line + "yes"
        else:
            line = line + "no"
        
    elif type(value) == list:   # lists
        # write out the current line
        line = line + "(0.." + repr(len(value) - 1) + ")"
        f.write(line)
        f.write("\n")
        line = ""

        # loop over elements in value printing out lines when
        # they reach > 70 characters or the next value would cause
        # the line to go over 80 characters
        for v in value:     
            if len(line) > 70:
                f.write(line)
                f.write("\n")
                line = ""

            to_add = repr(v)

            if len(line + " " + to_add) > 80:
                f.write(line)
                f.write("\n")
                line = ""

            if line != "":
                line = line + to_add+" "
            else:
                line = to_add + " "

    # write out the line and a newline character
    f.write(line)
    f.write("\n")

    return

# pulse program read/writing functions

def read_pprog(filename):
    """ 
    Read a Bruker pulse program (pulseprogram) file.

    Resultsing dictionary contains the following keys:
    
    ========    ===========================================================
    key         description
    ========    ===========================================================
    var         dictionary of variables assigned in pulseprogram
    incr        list of lists containing increment times
    loop        list of loop multipliers
    phase       list of lists containing phase elements
    ph_extra    list of lists containing comments at the end of phase lines
    ========    ===========================================================

    The incr,phase and ph_extra lists match up with loop list.  For example 
    incr[0],phase[0] and ph_extra[0] are all increment and phase commands 
    with comments which occur during loop 0 which has loop[0] steps.

    Parameters
    ----------
    filename : str
        Filename of pulseprogram file to read from,

    Returns
    -------
    dic : dict
        A dictionary with keys described above.
    
    See Also
    --------
    write_pprog : Write a Bruker pulse program to file.

    """
    
    # open the file
    f = open(filename, 'rb')

    # initilize lists and dictionaries
    var = dict()
    loop = []
    incr = [[]]
    phase = [[]]
    ph_extra = [[]]

    # loop over lines in pulseprogram looking for loops, increment, 
    # assigments and phase commands
    for line in f:

        # split line into comment and text and strip leading/trailing spaces
        if ";" in line:
            comment = line[line.index(";"):]
            text = line[:line.index(";")].strip()
        else:
            comment = ""
            text = line.strip()

        # remove label from text when first word is all digits or 
        # has "," as the last element
        if len(text.split()) != 0:
            s = text.split()[0]
            if s.isdigit() or s[-1] == ",":
                text = text[len(s):].strip()

        # skip blank lines and include lines
        if text == "" or text[0] == "#":
            #print line,"--Blank, Comment or Include"
            continue

        # see if we have quotes and have an assigment 
        # syntax "foo=bar"
        # add foo:bar to var dictionary
        if "\"" in text:
            if "=" in line:
                # strip quotes, split on = and add to var dictionary
                text = text.strip("\"")
                t = text.split("=")
                if len(t) >= 2:
                    key, value = t[0], t[1]
                    var[key] = value
                    #print line,"--Assignment"
                else:
                    pass
                    #print line,"--Statement"
                continue
            else:
                #print line,"--Statement"
                continue

        # loops begin with lo
        # syntax is: lo to N time M
        # add M to loop list
        if text[0:2] == "lo":
            loop.append(text.split()[4])
            incr.append([])
            phase.append([])
            ph_extra.append([])
            #print line,"--Loop"
            continue

        # increment statement have id, dd, ipu or dpu
        # syntax foo {id/dd/ipu/dpu}N
        # store N to incr list
        if ("id" in text) or ("dd" in text):
            incr[len(loop)].append(int(text.split()[1][2:]))
            #print line,"--Increment"
            continue

        if ("ipu" in text) or ("dpu" in text):
            incr[len(loop)].append(int(text.split()[1][3:]))
            #print line,"--Increment"
            continue

        # phase statement have ip or dp 
        # syntax fpp {ip/dp}N extra
        # store N to phase list and extra to ph_extra list
        if ("ip" in text) or ("dp" in text):
            _p = text.split()[1][2:]
            try:
                int_p=int(_p)
                phase[len(loop)].append(int_p)
            except:
                _p = _p.split('*')
                int_p = int(_p[0])
                for i in range(int(_p[1])):
                    phase[len(loop)].append(int_p)
            
            # find the first space after "ip" and read past there
            last = text.find(" ", text.index("ip"))
            if last == -1:
                ph_extra[len(loop)].append("")
            else:
                ph_extra[len(loop)].append(text[last:].strip())
            #print line,"--Phase"
            continue

        #print line,"--Unimportant"

    f.close()

    # remove the last empty incr, phase and ph_extra lists
    incr.pop()
    phase.pop()
    ph_extra.pop()

    # convert loop to numbers if possible
    for i, t in enumerate(loop):
        if t.isdigit():
            loop[i] = int(t)
        else:
            if var.has_key(t) and var[t].isdigit():
                loop[i] = int(var[t]) 

    # create the output dictionary
    dic = {"var":var, "incr":incr, "loop":loop, "phase":phase, 
            "ph_extra":ph_extra}

    return dic

def write_pprog(filename, dic, overwrite=False):
    """ 
    Write a minimal Bruker pulse program to file.
    
    **DO NOT TRY TO RUN THE RESULTING PULSE PROGRAM**

    This pulse program should return the same dictionary when read using 
    read_pprog, nothing else.  The pulse program will be nonsense.
    
    Parameters
    ----------
    filename : str
        Filename of file to write pulse program to.
    dic : dict
        Dictionary of pulse program parameters.
    overwrite : bool, optional
        True to overwrite an existing file, False will raise a Warning if the
        file already exists.

    See Also
    --------
    read_pprog : Read a Bruker pulse program.

    """
    
    # open the file for writing
    f = fileiobase.open_towrite(filename, overwrite=overwrite) 

    # write a comment
    f.write("; Minimal Bruker pulseprogram created by write_pprog\n")

    # write our the variables
    for k, v in dic["var"].iteritems():
        f.write("\"" + k + "=" + v + "\"\n")
    
    # write out each loop
    for i, steps in enumerate(dic["loop"]):
        
        # write our the increments
        for v in dic["incr"][i]:
            f.write("d01 id" + str(v) + "\n")
            
        # write out the phases
        for v, w in zip(dic["phase"][i], dic["ph_extra"][i]):
            f.write("d01 ip" + str(v) + " " + str(w) + "\n")

        f.write("lo to 0 times " + str(steps) + "\n")

    # close the file
    f.close()
        
    return

if __name__ == '__main__':
    import nose
    nose.main()


#! /usr/bin/env python
#===============================================================================
# utils
#===============================================================================
# Copyright (C) 2012-2013 Christian Fernandez
#       Laboratoire Catalyse et Spectrochimie, Caen, France.  
#       christian.fernandez@ensicaen.fr
# This software is governed by the CeCILL-B license under French law 
# and abiding by the rules of distribution of free software.  
# You can  use, modify and/ or redistribute the software under 
# the terms of the CeCILL-B license as circulated by CEA, CNRS and INRIA
# at the following URL "http://www.cecill.info".
# See Licence.txt in the main masai source directory
#===============================================================================
from .traitsutil import *
from .tools import *


# ******NOTICE***************
# optimize.py module by Travis E. Oliphant
#
# You may copy and use this module as you see fit with no
# guarantee implied provided you keep this notice in all copies.
# *****END NOTICE************

# Minimization routines
"""This module contain the fmin minimisation algorithm (SIMPLEX) used by Masai

It is adapted from the optimize.py module by Travis E. Oliphant 
(`Scipy package <http://docs.scipy.org/doc/scipy-0.7.x/reference/generated/scipy.optimize.fmin.html>`_)

"""
__all__ = ['fmin']

import numpy
from numpy import atleast_1d, sqrt, asfarray

def is_array_scalar(x):
    """Test whether `x` is either a scalar or an array scalar.

    """
    return len(atleast_1d(x) == 1)

import __builtin__
pymin = __builtin__.min
pymax = __builtin__.max
__version__="0.7"
_epsilon = sqrt(numpy.finfo(float).eps)


def wrap_function(function, args):
    ncalls = [0]
    def function_wrapper(x):
        ncalls[0] += 1
        return function(x, *args)
    return ncalls, function_wrapper

def fmin(func, x0, args=(), xtol=1e-4, ftol=1e-4, maxiter=None, maxfun=None,
         full_output=0, disp=1, retall=0, callback=None):
    """
    Minimize a function using the downhill simplex algorithm.

    This algorithm only uses function values, not derivatives or second
    derivatives.

    Parameters
    ----------
    func : callable func(x,*args)
        The objective function to be minimized.
    x0 : ndarray
        Initial guess.
    args : tuple
        Extra arguments passed to func, i.e. ``f(x,*args)``.
    callback : callable
        Called after each iteration, as callback(xk), where xk is the
        current parameter vector.

    Returns
    -------
    xopt : ndarray
        Parameter that minimizes function.
    fopt : float
        Value of function at minimum: ``fopt = func(xopt)``.
    iter : int
        Number of iterations performed.
    funcalls : int
        Number of function calls made.
    warnflag : string
        Message of warning or completion
    allvecs : list
        Solution at each iteration.

    Other parameters
    ----------------
    xtol : float
        Relative error in xopt acceptable for convergence.
    ftol : number
        Relative error in func(xopt) acceptable for convergence.
    maxiter : int
        Maximum number of iterations to perform.
    maxfun : number
        Maximum number of function evaluations to make.
    full_output : bool
        Set to True if fopt and warnflag outputs are desired.
    disp : bool
        Set to True to print convergence messages.
    retall : bool
        Set to True to return list of solutions at each iteration.

    Notes
    -----
    Uses a Nelder-Mead simplex algorithm to find the minimum of function of
    one or more variables.

    References
    ----------
    Nelder, J.A. and Mead, R. (1965), "A simplex method for function
    minimization", The Computer Journal, 7, pp. 308-313
    Wright, M.H. (1996), "Direct Search Methods: Once Scorned, Now
    Respectable", in Numerical Analysis 1995, Proceedings of the
    1995 Dundee Biennial Conference in Numerical Analysis, D.F.
    Griffiths and G.A. Watson (Eds.), Addison Wesley Longman,
    Harlow, UK, pp. 191-208.

    """
    fcalls, func = wrap_function(func, args)
    x0 = asfarray(x0).flatten()
    N = len(x0)
    rank = len(x0.shape)
    if not -1 < rank < 2:
        raise ValueError("Initial guess must be a scalar or rank-1 sequence.")
    if maxiter is None:
        maxiter = N * 200
    if maxfun is None:
        maxfun = N * 200

    rho = 1; chi = 2; psi = 0.5; sigma = 0.5;
    one2np1 = range(1,N+1)

    if rank == 0:
        sim = numpy.zeros((N+1,), dtype=x0.dtype)
    else:
        sim = numpy.zeros((N+1,N), dtype=x0.dtype)
    fsim = numpy.zeros((N+1,), float)
    sim[0] = x0
    if retall:
        allvecs = [sim[0]]
    fsim[0] = func(x0)
    nonzdelt = 0.05
    zdelt = 0.00025
    for k in range(0,N):
        y = numpy.array(x0,copy=True)
        if y[k] != 0:
            y[k] = (1+nonzdelt)*y[k]
        else:
            y[k] = zdelt

        sim[k+1] = y
        f = func(y)
        fsim[k+1] = f

    ind = numpy.argsort(fsim)
    fsim = numpy.take(fsim,ind,0)
    # sort so sim[0,:] has the lowest function value
    sim = numpy.take(sim,ind,0)

    iterations = 1
    abort = False
    while (fcalls[0] < maxfun and iterations < maxiter and not abort):
        if (max(numpy.ravel(abs(sim[1:]-sim[0]))) <= xtol \
            and max(abs(fsim[0]-fsim[1:])) <= ftol):
            break

        xbar = numpy.add.reduce(sim[:-1],0) / N
        xr = (1+rho)*xbar - rho*sim[-1]
        fxr = func(xr)
        doshrink = 0

        if fxr < fsim[0]:
            xe = (1+rho*chi)*xbar - rho*chi*sim[-1]
            fxe = func(xe)

            if fxe < fxr:
                sim[-1] = xe
                fsim[-1] = fxe
            else:
                sim[-1] = xr
                fsim[-1] = fxr
        else: # fsim[0] <= fxr
            if fxr < fsim[-2]:
                sim[-1] = xr
                fsim[-1] = fxr
            else: # fxr >= fsim[-2]
                # Perform contraction
                if fxr < fsim[-1]:
                    xc = (1+psi*rho)*xbar - psi*rho*sim[-1]
                    fxc = func(xc)

                    if fxc <= fxr:
                        sim[-1] = xc
                        fsim[-1] = fxc
                    else:
                        doshrink=1
                else:
                    # Perform an inside contraction
                    xcc = (1-psi)*xbar + psi*sim[-1]
                    fxcc = func(xcc)

                    if fxcc < fsim[-1]:
                        sim[-1] = xcc
                        fsim[-1] = fxcc
                    else:
                        doshrink = 1

                if doshrink:
                    for j in one2np1:
                        sim[j] = sim[0] + sigma*(sim[j] - sim[0])
                        fsim[j] = func(sim[j])

        ind = numpy.argsort(fsim)
        sim = numpy.take(sim,ind,0)
        fsim = numpy.take(fsim,ind,0)

        if callback is not None:
            abort = callback(sim[0], args[0], fcalls[0], iterations, min(fsim))
        iterations += 1
        if retall:
            allvecs.append(sim[0])

    x = sim[0]
    fval = min(fsim)
    warnflag = ""
    
    if abort:
        warnflag = "Warning: User termination - optimization was not finished"

    elif fcalls[0] >= maxfun:
        warnflag = "Warning: Maximum number of function evaluations has "\
                  "been exceeded."
    elif iterations >= maxiter:
        warnflag = "Warning: Maximum number of iterations has been exceeded"
    else:
        warnflag = "Optimization terminated successfully.\n"
        warnflag +="The returned value satisfies the termination criteria " \
                  "(using tolerance = %s)\n"% str(xtol)
        warnflag +="         Current function value: %f\n" % fval
        warnflag +="         Iterations: %d\n" % iterations
        warnflag +="         Function evaluations: %d\n" % fcalls[0]


    if full_output:
        retlist = x, fval, iterations, fcalls[0], warnflag
        if retall:
            retlist += (allvecs,)
    else:
        retlist = x
        if retall:
            retlist = (x, allvecs)

    return retlist

if __name__ == "__main__":
    pass



#! /usr/bin/env python
#===============================================================================
# utils
#===============================================================================
# Copyright (C) 2012-2013 Christian Fernandez
#       Laboratoire Catalyse et Spectrochimie, Caen, France.  
#       christian.fernandez@ensicaen.fr
# This software is governed by the CeCILL-B license under French law 
# and abiding by the rules of distribution of free software.  
# You can  use, modify and/ or redistribute the software under 
# the terms of the CeCILL-B license as circulated by CEA, CNRS and INRIA
# at the following URL "http://www.cecill.info".
# See Licence.txt in the main masai source directory
#===============================================================================
"""
.. _traitsutil:

This :mod:`traitsutil` module is containing some helper class used in other 
parts of the **MASAI** application. It is imported automatically when
loading the **masai.utils** package 

Examples
========

    >>> from masai.utils import InstanceUItem

"""

#===============================================================================
# Enthought imports
#===============================================================================
from traits.api import Str, Instance
from traitsui.api import UItem, InstanceEditor
from traits.api import Property, TraitFactory, TraitError
from traitsui.api import EnumEditor

#---------------------------------------------------------------------------
#  Utility functions:
#---------------------------------------------------------------------------

def super_getattr(object, name, *args):
    """Works the same way as getattr, except that name can be of the
    form 'a.b.c' (as many levels as you like).  For example:

    >>> class A:
    ... pass
    ...
    >>> a = A()
    >>> a.b = A()
    >>> a.b.c = 1
    >>> super_getattr(a, 'b.c')
    1
    >>> super_getattr(a.b, 'c')
    1
    """
    if '.' in name:
        attrs = name.split('.')
        last = attrs.pop()
        obj = object
        for attr in attrs:
            obj = getattr(obj, attr)
        return getattr(obj, last, *args)
    else:
        return getattr(object, name, *args)

def super_setattr(object, name, value):
    """Works the same way as setattr, except that name can be of the
    form 'a.b.c' (as many levels as you like).  For example:

    >>> class A:
    ... pass
    ...
    >>> a = A()
    >>> a.b = A()
    >>> super_setattr(a, 'b.c', 1)
    >>> a.b.c
    1
    """
    if '.' in name:
        attrs = name.split('.')
        last = attrs.pop()
        obj = object
        for attr in attrs:
            obj = getattr(obj, attr)
        setattr(obj, last, value)
    else:
        setattr(object, name, value)


#--------------------------------------------------------------------------------
# Helper class for DEnum trait.
#--------------------------------------------------------------------------------
class DEnumHelper(object):
    """Defines a bunch of staticmethods that collect all the helper
    functions needed for the DEnum trait.
    
    .. _denum:
    
    The :class:`DEnum` trait is a **dynamic enum** trait whose values are obtained from
    another trait on the object. (copied from mayavi.core.trait_def)

    .. warning::
       The problem with this trait is that the listeners (for changes to
       the valid values) are added only when the attribute is read or
       set.  Thus if the acceptable list of values are changed before the
       listeners are activated then the value will be set correctly only
       when it is accessed and not when the values are set.

    Written by: David C. Morrill and Prabhu Ramachandran
    (c) Copyright 2006-2008 by Enthought, Inc.


    """

    ######################################################################
    # Get/Set functions for the property.
    def get_value ( object, name ):
        return super_getattr(object, DEnumHelper._init_listeners(object, name))
    get_value = staticmethod(get_value)

    def set_value ( object, name, value ):
        _name = DEnumHelper._init_listeners( object, name )
        trait = object.trait( name )
        values = super_getattr(object, trait.values_name)
        if value not in values:
            raise TraitError, (object, name,
                               "one of %s"%values,
                               value )
        old = super_getattr(object, _name)
        super_setattr( object, _name, value )
        object.trait_property_changed(name, old, value)
    set_value = staticmethod(set_value)

    ######################################################################
    #  Makes a default EnumEditor for the trait:
    def make_editor ( trait = None ):
        return EnumEditor( name=trait.values_name )
    make_editor = staticmethod(make_editor)

    ######################################################################
    # Ensures that the listeners are initialized.
    def _init_listeners ( object, name ):
        _name = '_' + name
        if not hasattr( object, _name ):
            trait = object.trait( name )
            DEnumHelper._add_listeners( object, name, trait.values_name)
            default = trait.default or ''
            values = super_getattr( object, trait.values_name )
            if values:
                if default is None or default not in values:
                    default = values[0]
            super_setattr( object, _name, default )

        return _name
    _init_listeners = staticmethod(_init_listeners)

    def _add_listeners ( object, name, values_name ):
        def check_values(object, values_name, old, new):
            cur_choice = super_getattr(object, name)
            if cur_choice not in new:
                if new:
                    super_setattr(object, name, new[0])
                else:
                    super_setattr(object, name, '')

        def check_values_items(object, values_name, list_event):
            cur_choice = super_getattr(object, name)
            values = super_getattr(object, values_name[:-6])
            if cur_choice not in values:
                if values:
                    super_setattr(object, name, values[0])
                else:
                    super_setattr(object, name, '')

        object.on_trait_change( check_values,  values_name )
        object.on_trait_change( check_values_items, values_name + '_items' )
    _add_listeners = staticmethod(_add_listeners)


#-------------------------------------------------------------------------------
#  Defines the DEnum property:
#-------------------------------------------------------------------------------
DEnum = Property(DEnumHelper.get_value, DEnumHelper.set_value,
                 values_name = 'values',
                 editor  = (DEnumHelper.make_editor, {'trait': None})
                 )

DEnum = TraitFactory(DEnum)


#===============================================================================
# InstanceUItem
#===============================================================================
class InstanceUItem(UItem):
    """Convenience class for including an Instance in a View
    
    This class used as a workaround due to the impossibility to specyfy 
    the size of a group in a trait view as explained in the 
    traitsui/demo_group_size.py example of the *Enthougth EPD distribution*.
    
    """
    style = Str('custom')
    editor = Instance(InstanceEditor, ())

if __name__ == '__main__':
    pass